import logging
import os
from math import log10

from .constants import SPEED_OF_LIGHT


def lin_to_db(val):
    """Linear value to decibels

    Args:
        val : Linear scale value.

    Returns:
        Corresponding value in decibels.

    """
    return 10 * log10(val)


def db_to_lin(val_db):
    """Decibels to linear value

    Args:
        val : Value in decibels.

    Returns:
        Corresponding value in linear scale.

    """
    return 10**(val_db / 10)


def dbw_to_dbm(val_dbw):
    """Convert dBW to dBm

    Args:
        val_dbw : Value in dBW (decibels above 1 Watt).

    Returns:
        Value in dBm (decibels above 1 milliwatt).

    """
    return val_dbw + 30


def dbm_to_dbw(val_dbm):
    """Convert dBm to dBW

    Args:
        val_dbm : Value in dBm (decibels above 1 milliwatt).

    Returns:
        Value in dBW (decibels above 1 Watt).

    """
    return val_dbm - 30


def wavelength(freq):
    """Compute the radio wavelength for a given frequency

    Args:
        freq : Radio frequency in Hz.

    Returns:
        Wavelength in meters.

    """
    return SPEED_OF_LIGHT / freq


def _format_unit(value, thresholds, units):
    """Generate "value unit" string with an adequately scaled unit

    Select the highest unit that keeps the scaled value above unit. If none of
    the scaled values are above unit, output the lowest value (based on the
    lowest threshold).

    For example, consider the scaling of a bit rate of 1e4 bps with thresholds
    of 1e6 for Mbps, 1e3 for kbps, and 1 for bps. The implementation scans the
    thresholds in descending order, i.e., starting from 1e6, then 1e3, then
    1. In the given example, the value of 1e4 is equivalent to 0.01 Mbps, which
    is below unit (i.e., < 1). Hence, this first threshold is not suitable, and
    the implementation proceeds to the next. The next threshold is of 1e3
    (kbps) and results in 10 kbps. This value is above unit. Hence, the loop
    stops there and output 10 kbps.

    Args:
        value : Value to output as a string accompanied by the unit.
        thresholds : List of thresholds used to select the appropriate unit.
        units : List of strings with the units corresponding to each threshold.

    Returns:
        "value unit" string.

    """
    original_value = value
    zipped = zip(thresholds, units)
    for t, u in sorted(zipped, key=lambda t: t[0], reverse=True):
        value = original_value / t
        unit = u
        if (original_value >= t):
            break

    return "{:.2f} {}".format(value, unit)


def format_rate(rate):
    """Format data rate given in bps

    Args:
        rate : Data rate in bits per second (bps).

    Returns:
        String with the data rate and an adjusted unit such as Gbps, Mbps, or
        kbps.

    """
    return _format_unit(rate,
                        thresholds=[1e9, 1e6, 1e3, 1],
                        units=["Gbps", "Mbps", "kbps", "bps"])


def format_area(area):
    """Format area given in m2

    Args:
        area : Area in square meters (m2).

    Returns:
        String with the area and an adjusted unit such as cm2, dm2, or m2.

    """
    return _format_unit(area,
                        thresholds=[1, 1e-2, 1e-4],
                        units=["m2", "dm2", "cm2"])


def format_power(power):
    """Format power given in Watts

    Args:
        power : Power in Watts (W).

    Returns:
        String with the power value and an adjusted unit such as kW or mW.

    """
    return _format_unit(power,
                        thresholds=[1e3, 1, 1e-3],
                        units=["kW", "W", "mW"])


def log_header():
    """Log header to form a table with result logs"""
    w = 31  # width
    logging.info(w * "-" + "|" + w * "-")
    log_result("Parameter", "Value")
    logging.info(w * "-" + "|" + w * "-")


def log_result(parameter, value):
    """Log parameter-value result

    Args:
        parameter : Parameter to log.
        value     : Corresponding value.

    """
    logging.info("{:30s} | {:30s}".format(parameter, value))


def get_home_dir():
    """Get the user's home directory even if running with sudo"""
    sudo_user = os.environ.get('SUDO_USER')
    user = sudo_user if sudo_user is not None else ""
    return os.path.expanduser("~" + user)


def get_default_lb_dir():
    """Get the default directory for link-budget-specific files"""
    return os.path.join(get_home_dir(), ".link-budget")



"""
References:

 [1] Couch, Leon W.. Digital & Analog Communication Systems.
 [2] Lindgren, M. (2015). A 1296 MHz Earth–Moon–Earth Communication System
     (Master's thesis).
 [3] Timothy Pratt, Jeremy E. Allnutt, "Satellite Communications", 3rd Ed.
 [4] www.everythingrf.com/rf-calculators/parabolic-reflector-antenna-gain

"""

import unittest
from unittest.mock import Mock
from math import pi

from . import calc, util


class TestBudgetCalc(unittest.TestCase):

    def test_eirp(self):
        tx_power_dbw = 10
        tx_dish_gain_db = 17
        eirp_dbw = calc.eirp(tx_power_dbw, tx_dish_gain_db)
        self.assertEqual(eirp_dbw, 27)

    def test_carrier_eirp(self):
        # Example 6.2 from [3] for the case of Station A.
        tp_sat_power = 20
        carrier_power = calc.carrier_eirp(tp_sat_power,
                                          obo=3,
                                          peb=15e6,
                                          tp_bw=30e6)
        self.assertAlmostEqual(carrier_power, 14, places=1)
        # NOTE the carrier_eirp function takes the saturated EIRP, while the
        # above example considers the transponder output power only (before the
        # antenna). Nevertheless, the calculation is equivalent.

        # The transponder bandwidth must be provided for the PEB calculation:
        with self.assertRaises(ValueError):
            carrier_power = calc.carrier_eirp(tp_sat_power, obo=3, peb=15e6)

    def test_path_loss(self):
        # Example 4.2 from [3]
        self.assertAlmostEqual(
            calc.path_loss(d=40e6, freq=11e9),
            205.3,  # expected path loss
            places=1)

        # Examples based on Section 10.2 from [2]:
        distance = 364288e3
        freq = 1296e6
        rcs = 0.065 * 9.49e12

        # One-way path loss
        self.assertAlmostEqual(
            calc.path_loss(d=distance, freq=freq),
            206,  # expected one-way path loss
            delta=0.1)

        # Monostatic radar path loss
        radar_obj_gain_db = calc.radar_obj_gain(freq, rcs)
        self.assertAlmostEqual(
            calc.path_loss(d=distance,
                           freq=freq,
                           radar=True,
                           obj_gain=radar_obj_gain_db),
            270,  # expected two-way transmission loss
            delta=0.25)

        # Equivalent bistatic radar path loss (assuming the object-to-Rx
        # distance is equal to the Tx-to-object distance)
        self.assertAlmostEqual(
            calc.path_loss(d=distance,
                           freq=freq,
                           radar=True,
                           obj_gain=radar_obj_gain_db,
                           bistatic=True,
                           d_rx=distance),
            270,  # expected two-way transmission loss
            delta=0.25)

        # The gain of the radar object must be provided in radar mode
        with self.assertRaises(ValueError):
            calc.path_loss(d=distance, freq=freq, radar=True)

        # The Rx distance must be provided in bistatic radar mode
        with self.assertRaises(ValueError):
            calc.path_loss(d=distance,
                           freq=freq,
                           radar=True,
                           obj_gain=radar_obj_gain_db,
                           bistatic=True)

    def test_radar_object_gain(self):
        # Based on the example in Section 10.2 from [2]:
        freq = 1296e6
        rcs = 0.065 * 9.49e12
        Gobj = calc.radar_obj_gain(freq, rcs)
        self.assertAlmostEqual(Gobj, 141.7, delta=0.1)

    def test_rx_flux_density(self):
        """Test flux density computation"""
        # Example 4.1 from [3]
        distance = 4e7  # distance in meters
        eirp_dbw = 27  # EIRP in dBW
        Ae = 10  # effective aperture in m2

        # Flux density
        F_dbw_m2 = calc.rx_flux_density(eirp_dbw, distance)
        self.assertAlmostEqual(F_dbw_m2, -136, places=1)

        # Corresponding Rx power
        Ae_db = util.lin_to_db(Ae)  # effective aperture in dBm^2
        Pr_dbw = F_dbw_m2 + Ae_db
        self.assertAlmostEqual(Pr_dbw, -126, places=1)

        # This Rx power computation based on the flux density should match with
        # the Rx power computed using the antenna gain (i.e., based on the
        # "link equation"). Consider Example 4.2 from [3].
        freq = 11e9
        wavelength = util.wavelength(freq)
        rx_gain = util.lin_to_db(4 * pi * Ae / wavelength**2)
        path_loss = calc.path_loss(d=distance, freq=freq)
        Pr_dbw_2 = calc.rx_power(eirp_dbw, path_loss, rx_gain)
        self.assertAlmostEqual(Pr_dbw, Pr_dbw_2)

        # Check the match between the two Rx power computations (based on flux
        # density and based on the link equation) when a non-zero atmospheric
        # loss is introduced in the analysis.
        atm_loss_db = 0.5
        F_dbw_m2 = calc.rx_flux_density(eirp_dbw,
                                        distance,
                                        atm_loss_db=atm_loss_db)
        Pr_dbw = F_dbw_m2 + Ae_db
        Pr_dbw_2 = calc.rx_power(eirp_dbw,
                                 path_loss,
                                 rx_gain,
                                 atm_loss_db=atm_loss_db)
        self.assertAlmostEqual(Pr_dbw, Pr_dbw_2)

    def test_antenna_noise_temp(self):
        # Example 4.9 in [3]
        self.assertAlmostEqual(
            calc.antenna_noise_temp(attn_db=3.4),
            147,  # expected antenna noise temperature in K
            places=0)

    def test_coax_loss_nf(self):
        # Study aid example SA8-1 from [1]:
        loss_db, nf = calc.coax_loss_nf(length_ft=110, Tl=290)
        # At 290 K, the noise figure is equal to the line loss in dB:
        self.assertEqual(nf, 8.8)
        self.assertEqual(loss_db, 8.8)

    def test_cascaded_nf(self):
        # Study aid example SA8-1 from [1]:
        nfs = [0.6, 8.8, 10]
        gains = [40, -8.8]
        total_nf = calc.cascaded_noise_figure(nfs, gains)
        self.assertAlmostEqual(total_nf, 0.63, places=2)

        # If a single noise figure is provided, make sure the total noise
        # figure is equivalent to it.
        nf = 0.6
        total_nf = calc.cascaded_noise_figure(nfs=[nf], gains=[])
        self.assertEqual(total_nf, nf)

    def test_cascaded_input_noise_temp(self):
        # Example 4.3 from [3]:
        Tin = 25  # input from antenna
        Trf = 50  # RF amplifier
        Tm = 500  # mixer
        Tif = 1000  # IF amplifier
        Grf = 23  # RF amplifier gain
        Gm = -10  # Mixer gain (the mixer is lossy)
        temps = [Trf, Tm, Tif]
        gains = [Grf, Gm]
        Te = calc.cascaded_input_noise_temp(temps, gains)
        self.assertAlmostEqual(Te, 127.625 - Tin, places=3)

        # If a single input noise temperature is provided, the overall input
        # noise temperature must equivalent to it.
        temp = 50
        total_temp = calc.cascaded_input_noise_temp(temps=[temp], gains=[])
        self.assertEqual(total_temp, temp)

    def test_cascaded_nf_noise_temp_match(self):
        # The cascaded input noise temperature can either be obtained directly
        # or by first computing the cascaded noise figure, then converting the
        # noise figure to a noise temperature. The two results must match.
        temps = [50, 500, 1000]
        gains = [23, -10]
        Te = calc.cascaded_input_noise_temp(temps, gains)
        nfs = [calc.noise_temp_to_noise_fig(x) for x in temps]
        nf = calc.cascaded_noise_figure(nfs, gains)
        self.assertAlmostEqual(Te, calc.noise_fig_to_noise_temp(nf))

    def test_noise_fig_to_noise_temp(self):
        # Table 4.4 from [3]:
        noise_temp = [0, 20, 40]
        noise_fig = [0, 0.29, 0.56]
        for i, nf in enumerate(noise_fig):
            # Noise figure to noise temperature:
            Te = calc.noise_fig_to_noise_temp(nf)
            self.assertAlmostEqual(Te, noise_temp[i], delta=0.1)
            # Noise temperature to noise figure:
            self.assertAlmostEqual(calc.noise_temp_to_noise_fig(Te), nf)

    def test_rx_sys_noise_temp_lossless_feed(self):
        # Study aid example SA8-1 from [1]:
        Tsys = calc.rx_sys_noise_temp(Tar=20, Te=43.18)
        Tsys_dbk = util.lin_to_db(Tsys)
        self.assertAlmostEqual(Tsys_dbk, 18.01, places=2)

        # Example 4.3 from [3] for LNA gain of 50 dB and mixer gain of -10 dB:
        temps = [50, 500, 1000]
        gains = [50, -10]
        Tar = 25  # antenna output noise temperature
        Te = calc.cascaded_input_noise_temp(temps, gains)
        Tsys2 = calc.rx_sys_noise_temp(Tar, Te)
        self.assertAlmostEqual(Tsys2, 75.1, places=1)

    def test_rx_sys_noise_temp_lossy_feed(self):
        # Example 4.4 from [3]:
        temps = [50, 500, 1000]
        gains = [50, -10]
        waveguide_loss_db = 2
        waveguide_temp = 300
        Tar = 25  # antenna output noise temperature
        Te = calc.cascaded_input_noise_temp(temps, gains)
        Tsys = calc.rx_sys_noise_temp(Tar, Te, waveguide_loss_db,
                                      waveguide_temp)
        self.assertAlmostEqual(Tsys, 176.6, places=1)

    def test_spectral_density(self):
        # Double-sided power spectral density of the white noise produced by a
        # thermal noise source at 290 K:
        N0_over_2_dbm_hz = -174  # dBm/Hz
        N0_over_2_dbw_hz = util.dbm_to_dbw(N0_over_2_dbm_hz)  # dBW/Hz
        N0_over_2_w_hz = util.db_to_lin(N0_over_2_dbw_hz)  # Watts/Hz
        bw = 1e6  # Hz
        N = N0_over_2_w_hz * bw  # Watts
        N_dbw = util.lin_to_db(N)  # dBW
        self.assertEqual(N0_over_2_dbw_hz, calc.spectral_density(N_dbw, bw))

    def test_cnr(self):
        # Study aid example SA8-1 from [1]:
        C = calc.rx_power(eirp_db=52,
                          path_loss_db=205.73,
                          rx_ant_gain_db=32.96)
        N = calc.noise_power(T_sys_db=18.01, bw=24e6)
        cnr = calc.cnr(C, N)
        self.assertAlmostEqual(cnr, 16.03, places=1)

    def test_capacity(self):
        # At 0 dB SNR, the capacity in bits/complex-symbol becomes 1. The
        # corresponding capacity in bits per second (considering the symbol
        # rate) is equal to the passband (RF) bandwidth.
        self.assertEqual(
            calc.capacity(snr_db=0, bw=1e3),
            1e3  # expected capacity in bps
        )

    def test_carrier_to_asi_ratio(self):
        # Test the 45 cm dish example from Rec. ITU-R BO.1213-1, where the
        # on-axis antenna gain is 33.3 dB and the off-axis gain is roughly 30
        # dB at a 2° off-axis angle. Note the off-axis gain model adopted in
        # the Antenna object comes from Rec. ITU-R S.465-6, but the main lobe
        # (including the 2° angle) comes from APEREC026V01 and ITU-R BO.1213-1.
        rx_dish = Mock()
        rx_dish.gain_db = 33.3
        rx_dish.off_axis_gain.return_value = 30
        long_separation = 2
        # If the adjacent satellites transmit with the same EIRP as the wanted
        # satellite, the C/I becomes equal to the difference between the
        # on-axis and off-axis antenna gains, which is of 3.3 dB:
        c_to_i_db = calc.carrier_to_asi_ratio(rx_dish,
                                              long_separation,
                                              asi_eirp_ratio=1)
        self.assertAlmostEqual(c_to_i_db, 3.3, places=1)

        # If the adjacent satellites transmit with an aggregate EIRP twice as
        # stronger as the wanted EIRP, the C/I reduces by 3 dB:
        c_to_i_db = calc.carrier_to_asi_ratio(rx_dish,
                                              long_separation,
                                              asi_eirp_ratio=2.0)
        self.assertAlmostEqual(c_to_i_db, 0.3, places=1)

        # Lastly, in the opposite case, if the adjacent satellites collectively
        # transmit with half the wanted EIRP, the C/I increases by 3 dB:
        c_to_i_db = calc.carrier_to_asi_ratio(rx_dish,
                                              long_separation,
                                              asi_eirp_ratio=0.5)
        self.assertAlmostEqual(c_to_i_db, 6.3, places=1)

    def test_cnir(self):
        # Example 4.10.b from [3]:
        cnr_db = 17
        ci_db = 24
        self.assertAlmostEqual(calc.cnir(cnr_db, ci_db), 16.2, places=1)



"""
Antenna Calculations

References:

- [1] Timothy Pratt, Jeremy E. Allnutt, "Satellite Communications", 3rd Ed.
- [2] Couch, Leon W.. Digital & Analog Communication Systems.
- [3] Recommendation ITU-R S.465-6.

"""
import logging
from math import log10, pi, sqrt

from . import util


class Antenna:
    """Parabolic antenna

    Args:
        freq (float): Operating frequency in Hz.
        gain (float): Antenna gain in dB.
        diameter (float): Diameter in m.
        efficiency (float): Aperture efficiency.
        label (str): Label used for logs.

    Note:
        According to [1], the aperture efficiency is typically in the range
        0.5–0.75 for parabolodial reflector antennas, lower for small antennas
        and higher for large Cassegrain and Gregorian antennas. For instance,
        Table 8-4 in [2] assumes an aperture efficiency of 0.56.

    Attributes:
        effective_aperture : Effective aperture area.
        gain_db : Antenna gain in dBi.

    """

    def __init__(self,
                 freq,
                 gain=None,
                 diameter=None,
                 efficiency=None,
                 label="Dish"):
        """Construct the antenna object"""
        if (freq is None):
            raise ValueError("Operating frequency is required")

        able_to_calc_gain = diameter is not None and efficiency is not None
        if (gain is None and not able_to_calc_gain):
            raise ValueError("gain must be provided when the diameter, freq, "
                             "and efficiency parameters are not")

        self.freq = freq
        self.diameter = diameter
        self.aperture_efficiency = efficiency

        # When the gain is explicitly provided, infer the effective aperture
        # area from it. Otherwise, compute the effective aperture from the
        # given antenna diameter and aperture efficiency. Then, using the
        # computed effective aperture, compute the antenna gain.
        if (gain is None):
            self.effective_aperture = self._calc_eff_aperture(
                diameter, efficiency)
            self.gain_db = self._calc_gain(freq, self.effective_aperture)
            util.log_result("{} aperture efficiency".format(label),
                            "{:.1f} %".format(efficiency * 100))
        else:
            self.gain_db = gain
            self.effective_aperture = self._infer_eff_aperture(freq, gain)
            # If the aperture efficiency is provided, infer the physical
            # diameter of an equivalent parabolic reflector.
            if (efficiency is not None):
                self.diameter = self._infer_diameter(self.effective_aperture,
                                                     self.aperture_efficiency)
                util.log_result("{} inferred diameter".format(label),
                                "{:.2f} m".format(self.diameter))
            else:
                logging.warning(
                    "Cannot infer the physical diameter of {} "
                    "(aperture efficiency required).".format(label))

        util.log_result("{} gain".format(label),
                        "{:.2f} dB".format(self.gain_db))
        util.log_result("{} effective aperture".format(label),
                        "{}".format(util.format_area(self.effective_aperture)))

    def _calc_eff_aperture(self, diameter, aperture_efficiency):
        """Compute the antenna's effective aperture area

        The effective aperture area is given by:

        .. math::

          A_e = \\eta A,

        where :math:`A` represents the antenna's physical aperture area and
        :math:`\\eta` is the aperture efficiency.

        If the aperture is circular with a diameter :math:`D` in meters (or
        radius :math:`r`), the physical aperture area is given by:

        .. math::

          A = 𝜋 r^2 = \\frac{𝜋 D^2}{4}.

        Hence, the effective aperture area becomes:

        .. math::

          A_e = \\frac{\\eta 𝜋 D^2}{4}.

        Args:
            diameter (float): Diameter in m.
            aperture_efficiency (float): Aperture efficiency.

        Returns:
            Effective aperture area in square meters (m^2).

        """
        radius = diameter / 2
        face_area = pi * (radius**2)  # assume circular area
        return face_area * aperture_efficiency

    def _calc_gain(self, freq, effective_aperture):
        """Calculate the parabolic dish gain

        The gain in linear units is given by:

        .. math::

          G = \\frac{4𝜋A_e}{𝜆^2},

        where :math:`A_e` is the effective aperture and 𝜆 is the wavelength.

        Args:
            freq (float): Operating frequency in Hz.
            effective_aperture (float): Effective aperture in square meters.

        Returns:
            Gain in dB.

        """
        wavelength = util.wavelength(freq)
        gain = effective_aperture * 4 * pi / (wavelength**2)
        return util.lin_to_db(gain)

    def _infer_eff_aperture(self, freq, gain_db):
        """Infer the effective aperture area from the antenna gain

        The effective aperture area can be inferred from the gain and
        wavelength, as follows:

        .. math::

          A_e = \\frac{G 𝜆^2}{4𝜋}.

        Args:
            freq (float): Operating frequency in Hz.
            gain_db (float): Antenna gain in dB.

        Returns:
            float: Effective aperture area in square meters (m^2).

        """
        wavelength = util.wavelength(freq)
        gain = util.db_to_lin(gain_db)
        return gain * (wavelength**2) / (4 * pi)

    def _infer_diameter(self, effective_aperture, aperture_efficiency):
        """Infer the diameter of an equivalent parabolic reflector

        The physical aperture area :math:`A` can be expressed in terms of the
        effective aperture :math:`A_e` and the aperture efficiency
        :math:`\\eta`, as follows:

        .. math::

          A = \\frac{A_e}{\\eta}.

        If the aperture is circular with a diameter D in meters, the physical
        aperture area is given by:

        .. math::

          A = \\frac{𝜋 D^2}{4}.

        Hence, it follows that:

        .. math::

          D = \\sqrt{\\frac{4 A_e}{\\eta \\pi}}.

        Note:
            This inference is useful when working with a non-parabolic antenna,
            such as a flat-panel antenna. In this case, you may know the
            antenna gain and aperture efficiency specifications, but not the
            physical diameter and aperture of the antenna. Meanwhile, the
            diameter may still be required for computations such as the
            tropospheric scintillation model from Recommendation ITU-R 618
            (see, e.g., `ITU-Rpy
            <https://itu-rpy.readthedocs.io/en/latest/apidoc/itu618.html>`_).

        Args:
            effective_aperture (float): Effective aperture in square meters.
            aperture_efficiency (float): Aperture efficiency.

        Returns:
            float: Diameter in meters (m) of an equivalent parabolic reflector
            with the same aperture efficiency.

        """
        return sqrt((4 * effective_aperture) / (aperture_efficiency * pi))

    def off_axis_gain(self, angle):
        """Compute the off-axis co-polar antenna gain

        Based on the reference earth station radiation pattern from ITU-R
        S.465-6 (01/2010 version) and the APEREC026V01 standard from the ITU-R
        antenna pattern list in
        https://www.itu.int/en/ITU-R/software/Pages/ant-pattern.aspx.

        Args:
            angle (float): Off-axis angle in degrees relative to the boresight.
                Must be within the [0, 180°) range.

        Returns:
            (float) Off-axis antenna gain in dBi.

        """
        if self.freq < 2e9 or self.freq > 31e9:
            raise ValueError(
                "Off-axis gain model only valid for frequency in [2, 31] GHz")
        D_over_lambda = self.diameter / util.wavelength(self.freq)

        if (D_over_lambda < 33.3):
            phi_min = 2.5  # See NOTE 5 in ITU-R S.465-6.
        elif (D_over_lambda >= 50):
            phi_min = max(1, 100 / D_over_lambda)
        else:
            phi_min = max(2, 114 * D_over_lambda**(-1.09))

        Gmax = self.gain_db

        if (angle >= 0 and angle < phi_min):
            # Recommendation ITU-R S.465-6 does not define the radiation
            # pattern for this angle range (corresponding to the main lobe).
            # However, other models do. For instance, APEREC026V01 from
            # https://www.itu.int/en/ITU-R/software/Pages/ant-pattern.aspx
            # (also based on ITU-R S.465-6) defines the following gain:
            G = Gmax - 2.5e-3 * (D_over_lambda * angle)**2

            # The above is the same gain adopted in other recommendations, such
            # as Rec. ITU-R BO.1213-1. However, APEREC026V01 defines various
            # exceptions depending on the D/lambda metric, as follows:
            if (D_over_lambda >= 33.3 and D_over_lambda <= 54.5):
                phi_1 = 0.9 * 114 * (D_over_lambda**(-1.09))
                if (angle >= phi_1):
                    G = max(G, 32 - 25 * log10(angle))
            elif (D_over_lambda > 54.5):
                phi_r = 15.85 * (D_over_lambda**(-0.6))
                G1 = 32 - 25 * log10(phi_r)
                phi_m = (20 / D_over_lambda) * sqrt(Gmax - G1)
                if (angle >= phi_m and angle <= phi_r):
                    G = G1
                elif (angle > phi_r):
                    G = max(32 - 25 * log10(angle), -10)
            # NOTE that APEREC026V01 does not use the phi_min limit at all if
            # D/lambda > 54.5. However, we prioritize the model from ITU-R
            # S.465-6, which uses and defines phi_min for all D/lambda.
        elif (angle >= phi_min and angle < 48):
            G = max(32 - 25 * log10(angle), -10)
        elif (angle >= 48 and angle <= 180):
            G = -10
        else:
            raise ValueError(
                "Off-axis angle {} is out of range.".format(angle))

        return G



"""Link budget analysis unit tests

References:
 [1] Couch, Leon W.. Digital & Analog Communication Systems.
 [2] Lindgren, M. (2015). A 1296 MHz Earth–Moon–Earth Communication System
     (Master's thesis).
 [3] Timothy Pratt, Jeremy E. Allnutt, "Satellite Communications", 3rd Ed.

"""
import copy
import io
import json
import sys
import unittest
import unittest.mock
from itertools import combinations
from . import cli


class TestCli(unittest.TestCase):

    def setUp(self):
        # Arguments to configure Example SA8-1 from [1]:
        self.base_args = [
            '--eirp', '52', '--freq', '12.45e9', '--if-bw', '24e6',
            '--rx-dish-size', '0.46', '--rx-dish-efficiency', '0.557',
            '--atmospheric-loss', '0', '--antenna-noise-temp', '20',
            '--lnb-noise-fig', '0.6', '--lnb-gain', '40', '--coax-length',
            '110', '--rx-noise-fig', '10', '--sat-long', '-101', '--rx-long',
            '-82.43', '--rx-lat', '29.71'
        ]

    def test_ku_band_example(self):
        # Example SA8-1 from [1]:
        parser = cli.get_parser()
        args = parser.parse_args(self.base_args)
        cli.validate(parser, args)
        res = cli.analyze(args)
        self.assertAlmostEqual(res['cnr_db'], 15.95, places=2)
        # The actual result in [1] is distinct due to various roundings.

    def test_c_band_rain_example(self):
        # Example 4.6 in [3] (clear air):
        parser = cli.get_parser()
        args = parser.parse_args([
            '--eirp',
            '34',  # 39 -2 dB backoff -3 dB beam edge contour
            '--freq',
            '4e9',
            '--if-bw',
            '30e6',
            '--rx-dish-gain',
            '49.7',
            '--mispointing-loss',
            '0.5',
            '--atmospheric-loss',
            '0.2',
            '--lnb-noise-temp',
            '45',
            '--lnb-gain',
            '60',  # not given - consider a high gain
            '--coax-length',
            '0',
            '--rx-noise-fig',
            '0',  # LNA noise only
            '--slant-range',
            '40e3'
        ])
        cli.validate(parser, args)
        res = cli.analyze(args, verbose=True)
        self.assertAlmostEqual(res['cnr_db'], 22.7, places=1)

        # Example 4.7 in [3] (same as Example 4.6, but with heavy rain):
        args = parser.parse_args([
            '--eirp',
            '34',  # 39 -2 dB backoff -3 dB beam edge contour
            '--freq',
            '4e9',
            '--if-bw',
            '30e6',
            '--rx-dish-gain',
            '49.7',
            '--mispointing-loss',
            '0.5',
            '--atmospheric-loss',
            '1.2',  # 1 dB rain attenuation
            '--lnb-noise-temp',
            '45',
            '--lnb-gain',
            '60',  # not given - consider a high gain
            '--coax-length',
            '0',
            '--rx-noise-fig',
            '0',  # LNA noise only
            '--slant-range',
            '40e3'
        ])
        cli.validate(parser, args)
        res = cli.analyze(args, verbose=True)
        # The book seems to have an error on the clear air noise power given in
        # Table 4.5b, which is given as −135.5 dBW instead of −136.2 dBW as
        # given in Table 4.5a. That is, it is considering a noise power 0.7 dBW
        # higher. In the end, the book considers a CNR of 18.2 dB in heavy
        # rain, but we assume the correct one is 18.9 dBW.
        self.assertAlmostEqual(res['cnr_db'], 18.9, places=1)

    def test_radar_example(self):
        # Based on [2]. Most of the calculation is in Chapters 10 and 11. Other
        # important sections are highlighted below.
        parser = cli.get_parser()
        radar_cross_section = 0.065 * 9.49e12  # See Section 10.2
        args = parser.parse_args([
            '--eirp',
            '55.6',
            '--freq',
            '1296e6',
            '--if-bw',
            '100',
            '--rx-dish-gain',
            '31.1',  # See Section 8.7.2
            '--atmospheric-loss',  # Neglected (See Section 3.5.6)
            '0',
            '--antenna-noise-temp',
            '51.8',  # See Section 8.10
            '--lnb-noise-fig',
            '0.52',  # Chosen to achieve a total NF of 0.54
            # dB, as in Table 8.4
            '--lnb-gain',
            '36.2',  # LNA gain in Table 8.4
            '--coax-length',
            '32.8',  # 32.8 ft ~= 10 m
            '--rx-noise-fig',
            '10',  # See Table 8.4
            '--slant-range',
            '364288',  # See Section 10.2
            '--radar',
            '--radar-cross-section',
            str(radar_cross_section)
        ])
        cli.validate(parser, args)
        res = cli.analyze(args)
        self.assertAlmostEqual(res['cnr_db'], 5.51, places=2)
        # The final result differs slightly from the one presented in [2] due
        # to various rounding of intermediate results.

    def test_ku_band_example_with_tle_sat_pos(self):
        """Test Ku band example with satellite position given by TLE"""
        # Example SA8-1 from [1]:
        parser = cli.get_parser()
        args = parser.parse_args([
            '--eirp', '52', '--freq', '12.45e9', '--if-bw', '24e6',
            '--rx-dish-size', '0.46', '--rx-dish-efficiency', '0.557',
            '--atmospheric-loss', '0', '--antenna-noise-temp', '20',
            '--lnb-noise-fig', '0.6', '--lnb-gain', '40', '--coax-length',
            '110', '--rx-noise-fig', '10', '--rx-long', '-82.43', '--rx-lat',
            '29.71', '--sat-tle-name', 'Directv 9S'
        ])
        cli.validate(parser, args)
        res = cli.analyze(args)
        self.assertAlmostEqual(res['cnr_db'], 15.95, delta=0.01)

        # If the TLE name is invalid, the analyze function exits
        with self.assertRaises(SystemExit):
            args.sat_tle_name = 'Invalid Name'
            res = cli.analyze(args)

    def test_satellite_below_horizon(self):
        """Test satellite below horizon from the observer location"""
        parser = cli.get_parser()
        args = parser.parse_args([
            '--eirp', '52', '--freq', '12.45e9', '--if-bw', '24e6',
            '--rx-dish-size', '0.46', '--rx-dish-efficiency', '0.557',
            '--atmospheric-loss', '0', '--antenna-noise-temp', '20',
            '--lnb-noise-fig', '0.6', '--lnb-gain', '40', '--coax-length',
            '110', '--rx-noise-fig', '10', '--rx-long', '-82.43', '--rx-lat',
            '29.71', '--sat-tle-name', 'TELSTAR 18V (APSTAR 5C)'
        ])
        cli.validate(parser, args)
        res = cli.analyze(args)
        self.assertIsNone(res)

    def test_ku_band_example_with_atmospheric_attn(self):
        """Test Ku band example while considering atmospheric attenuation"""
        # Same arguments based on Example SA8-1 from [1], but now with the
        # atmospheric attenuation omitted such that the tool computes it
        # automatically from models. The antenna noise temperature is also
        # omitted such that it can be derived from the atmospheric
        # attenuation. Assume also circular polarization to hit that condition
        # on code coverage.
        availability = 99.99
        args = [
            '--eirp', '52', '--freq', '12.45e9', '--if-bw', '24e6',
            '--polarization', 'circular', '--availability',
            str(availability), '--rx-dish-size', '0.46',
            '--rx-dish-efficiency', '0.557', '--lnb-noise-fig', '0.6',
            '--lnb-gain', '40', '--coax-length', '110', '--rx-noise-fig', '10',
            '--sat-long', '-101', '--rx-long', '-82.43', '--rx-lat', '29.71'
        ]
        parser = cli.get_parser()
        args = parser.parse_args(args)
        cli.validate(parser, args)
        res = cli.analyze(args)

        # With circular polarization, the LNB polarization skew should be
        # assumed equal to 45 degrees.
        self.assertEqual(res['pointing']['polarization_skew'], 45)

        # The atmospheric attenuation modeled for a high link availability
        # (99.99%) in Ku band is usually significant.
        expected_exceeded_attn_db = 8
        self.assertGreater(res['atmospheric_loss_db'],
                           expected_exceeded_attn_db)

        # The resulting C/N is expected to be lower than in
        # test_ku_band_example (which assumed zero atmospheric loss).
        self.assertLess(res['cnr_db'], 15.95 - expected_exceeded_attn_db)

        # TODO: replace this test by an example that can be referenced to a
        # publication. The current version is only guessing a reasonable
        # attenuation level so that the atmospheric modeling can be covered.

    def test_ku_band_example_with_margin(self):
        """Test Ku band example with the link margin for a target min CNR

        Introduce a hypothetical minimum CNR and implementation margin in the
        scenario of Example SA8-1 from [1].

        """
        min_cnr = 2
        impl_margin = 1
        parser = cli.get_parser()
        args = parser.parse_args(
            self.base_args +
            ['--min-cnr',
             str(min_cnr), '--impl-margin',
             str(impl_margin)])
        cli.validate(parser, args)
        res = cli.analyze(args)
        self.assertAlmostEqual(res['cnr_db'], 15.95, places=2)
        self.assertAlmostEqual(res['margin_db'],
                               15.95 - (min_cnr + impl_margin),
                               places=2)

    def test_ku_band_example_with_asi(self):
        """Test Ku band scenario with adjacent satellite interference (ASI)

        According to ITU-R BO.1213-1, the difference between the on-axis and
        off-axis gain at 2° with a 45 cm dish of 0.65 aperture efficiency is
        roughly 3.3 dB when operating at 12.2 GHz. Hence, the ASI (or C/I) will
        dominate the carrier to noise plus interference ratio (CNIR). In this
        case, the C/(N+I) has to be less than the C/I of 3.3 dB.

        """
        base_args = [
            '--eirp', '52', '--freq', '12.2e9', '--if-bw', '24e6',
            '--rx-dish-size', '0.45', '--rx-dish-efficiency', '0.65',
            '--atmospheric-loss', '0', '--antenna-noise-temp', '20',
            '--lnb-noise-fig', '0.6', '--lnb-gain', '40', '--coax-length',
            '110', '--rx-noise-fig', '10', '--sat-long', '-101', '--rx-long',
            '-82.43', '--rx-lat', '29.71', '--min-cnr', '1'
        ]
        parser = cli.get_parser()
        args = parser.parse_args(base_args)

        # Without ASI, the C/(N+I) (equal to the C/N) is more than 16.4 dB
        cli.validate(parser, args)
        res = cli.analyze(args)
        self.assertEqual(res['cnr_db'], res['cnir_db'])
        self.assertGreater(res['cnir_db'], 16.4)
        no_asi_cnir_db = res['cnir_db']
        no_asi_capacity_bps = res['capacity_bps']
        no_asi_margin_db = res['margin_db']

        # With ASI from an adjacent satellite 2° away, the C/(N+I) drops
        # drastically to a value close to the C/I
        parser = cli.get_parser()
        args = parser.parse_args(base_args +
                                 ['--asi', '--asi-long-separation',
                                  str(2)])
        cli.validate(parser, args)
        res = cli.analyze(args)
        self.assertAlmostEqual(res['ci_db'], 3.35, places=1)
        self.assertLess(res['cnir_db'], res['ci_db'])

        # The C/N should remain the same as the C/(N+I) from the evaluation
        # without ASI (where I=0):
        self.assertEqual(res['cnr_db'], no_asi_cnir_db)

        # The capacity and link margin should both consider the C/(N+I), not
        # the C/N, so they should have reduced:
        self.assertLess(res['capacity_bps'], no_asi_capacity_bps)
        self.assertLess(res['margin_db'], no_asi_margin_db)

    def test_opts(self):
        """Test program options"""
        parser = cli.get_parser()

        # Base parameters: same as the ones adopted in "test_ku_band_example"
        base_args = [
            '--freq', '12.45e9', '--if-bw', '24e6', '--rx-dish-size', '0.46',
            '--rx-dish-efficiency', '0.557', '--atmospheric-loss', '0',
            '--antenna-noise-temp', '20', '--lnb-noise-fig', '0.6',
            '--lnb-gain', '40', '--coax-length', '110', '--rx-noise-fig', '10',
            '--sat-long', '-101'
        ]

        # Test out-of-range latitude and longitude coordinates
        for rx_lat in [-91, 91]:
            with self.assertRaises(SystemExit):
                args = parser.parse_args(base_args + [
                    '--eirp', '52', '--sat-long', '-101', '--rx-long',
                    '-82.43', '--rx-lat',
                    str(rx_lat)
                ])
                cli.validate(parser, args)
        for rx_long in [-181, 181]:
            with self.assertRaises(SystemExit):
                args = parser.parse_args(base_args + [
                    '--eirp', '52', '--sat-long', '-101', '--rx-long',
                    str(rx_long), '--rx-lat', '29.71'
                ])
                cli.validate(parser, args)
        for sat_long in [-181, 181]:
            with self.assertRaises(SystemExit):
                args = parser.parse_args(base_args + [
                    '--eirp', '52', '--sat-long',
                    str(sat_long), '--rx-long', '-82.43', '--rx-lat', '29.71'
                ])
                cli.validate(parser, args)

        # Now with the valid coordinates adopted in "test_ku_band_example"
        base_args.extend(
            ['--sat-long', '-101', '--rx-long', '-82.43', '--rx-lat', '29.71'])

        # Set the EIRP of 52 dBW indirectly through the Tx power and dish gain
        args = parser.parse_args(base_args +
                                 ['--tx-power', '20', '--tx-dish-gain', '32'])
        cli.validate(parser, args)
        res = cli.analyze(args)
        self.assertAlmostEqual(res['cnr_db'], 15.95, places=2)

        # Set the EIRP based on the Tx power and the Tx dish size. Consider a
        # Tx power of 5 dBW, and a 2.4m dish with 55.7% aperture efficiency,
        # which should lead to an EIRP of 52.37 dBW
        args = parser.parse_args(base_args + [
            '--tx-power', '5', '--tx-dish-size', '2.4', '--tx-dish-efficiency',
            '0.557'
        ])
        cli.validate(parser, args)
        res = cli.analyze(args)
        self.assertAlmostEqual(res['cnr_db'], 15.95 + 0.37, places=2)

        # Tx power without the Tx dish gain or size should throw error:
        with self.assertRaises(SystemExit):
            args = parser.parse_args(base_args + ['--tx-power', '20'])
            cli.validate(parser, args)

        # Provide the carrier PEB but not the transponder BW
        with self.assertRaises(SystemExit):
            args = parser.parse_args(base_args +
                                     ['--eirp', '52', '--carrier-peb', '1e6'])
            cli.validate(parser, args)

        # Rx dish gain given directly instead of through the dish size
        base_args = [
            '--eirp', '52', '--freq', '12.45e9', '--if-bw', '24e6',
            '--atmospheric-loss', '0', '--antenna-noise-temp', '20',
            '--lnb-noise-fig', '0.6', '--lnb-gain', '40', '--coax-length',
            '110', '--rx-noise-fig', '10', '--sat-long', '-101', '--rx-long',
            '-82.43', '--rx-lat', '29.71'
        ]
        args = parser.parse_args(base_args + ['--rx-dish-gain', '33.024'])
        cli.validate(parser, args)
        res = cli.analyze(args)
        self.assertAlmostEqual(res['cnr_db'], 15.95, places=2)

        # LNB noise temperature instead of the LNB noise figure
        base_args = [
            '--eirp', '52', '--freq', '12.45e9', '--if-bw', '24e6',
            '--rx-dish-size', '0.46', '--rx-dish-efficiency', '0.557',
            '--atmospheric-loss', '0', '--antenna-noise-temp', '20',
            '--lnb-noise-temp', '42.964', '--lnb-gain', '40', '--coax-length',
            '110', '--rx-noise-fig', '10', '--sat-long', '-101', '--rx-long',
            '-82.43', '--rx-lat', '29.71'
        ]
        args = parser.parse_args(base_args)
        res = cli.analyze(args)
        self.assertAlmostEqual(res['cnr_db'], 15.95, places=2)

        # Radar mode requires the radar object's cross section
        with self.assertRaises(SystemExit):
            args = parser.parse_args(base_args + ['--radar'])
            cli.validate(parser, args)

        # Slant range is mutual exclusive with satellite and Rx coordinates
        with self.assertRaises(SystemExit):
            args = parser.parse_args(base_args + ['--slant-range', 0])
            cli.validate(parser, args)

        # If the slant is not provided, the satellite and Rx station
        # coordinates are required
        base_args = [
            '--eirp', '52', '--freq', '12.45e9', '--if-bw', '24e6',
            '--rx-dish-size', '0.46', '--antenna-noise-temp', '20',
            '--lnb-noise-fig', '0.6', '--lnb-gain', '40', '--coax-length',
            '110', '--rx-noise-fig', '10'
        ]
        pos_args = [['--sat-long', '-101'], ['--rx-long', '-82.43'],
                    ['--rx-lat', '29.71']]
        # Test all combinations of one or two Rx/sat position args, all of
        # which should fail (all the three args are required)
        for n_comb in [1, 2]:
            for selected_pos_args in list(combinations(pos_args, n_comb)):
                with self.assertRaises(SystemExit):
                    test_args = copy.copy(base_args)
                    for pos_arg_pair in selected_pos_args:
                        test_args.extend(pos_arg_pair)
                    args = parser.parse_args(test_args)
                    cli.validate(parser, args)

        # When modeling the atmospheric loss, the satellite and earth station
        # coordinates are required.
        with self.assertLogs(level='ERROR'):  # An error message is logged.
            with self.assertRaises(SystemExit):
                args = parser.parse_args(base_args + ['--slant-range', 0])
                cli.validate(parser, args)

        # Also, the link availability must be within [95, 99.999].
        base_args += [
            '--sat-long', '-101', '--rx-long', '-82.43', '--rx-lat', '29.71'
        ]
        for availability in [50, 94.999, 99.9999, 100]:
            with self.assertRaises(SystemExit):
                args = parser.parse_args(
                    base_args +
                    ['--availability', str(availability)])
                cli.validate(parser, args)

        # A warning is printed if the --antenna-noise-temp option is given and
        # --atmospheric-loss is not.
        with self.assertLogs(level='WARNING'):
            args = parser.parse_args(base_args)
            cli.validate(parser, args)

        # Try a negative implementation margin
        with self.assertRaises(SystemExit):
            args = parser.parse_args(self.base_args + ['--impl-margin', '-1'])
            cli.validate(parser, args)

    @unittest.mock.patch('sys.stdout', new_callable=io.StringIO)
    def test_json_verbose_output(self, mock_stdout):
        """Test the output when the analyzer runs in verbose and --json mode"""
        parser = cli.get_parser()
        args = parser.parse_args(self.base_args + ['--json'])
        # Run the analyzer in verbose mode, in which case it should print the
        # JSON results to stdout in addition to returning them.
        cli.validate(parser, args)
        json_res = cli.analyze(args, verbose=True)
        printed_json = json.loads(mock_stdout.getvalue())
        self.assertEqual(printed_json, json_res)

    @unittest.mock.patch('sys.stdout', new_callable=io.StringIO)
    def test_main_entrypoint_json(self, mock_stdout):
        """Test the main program entrypoint with json output"""

        # Place arguments on argv
        old_sys_argv = sys.argv
        sys.argv = [old_sys_argv[0]] + self.base_args + ['--json']

        # Run the main entrypoint
        try:
            cli.main()
        finally:
            sys.argv = old_sys_argv

        # Check the JSON results printed to stdout
        printed_json = json.loads(mock_stdout.getvalue())
        self.assertAlmostEqual(printed_json['cnr_db'], 15.95, places=2)

    def test_main_entrypoint(self):
        """Test the main program entrypoint with normal (non-json) output"""

        # Place arguments on argv
        old_sys_argv = sys.argv
        sys.argv = [old_sys_argv[0]] + self.base_args

        # Run the main entrypoint and capture the info logs
        with self.assertLogs(level='INFO') as cm:
            try:
                cli.main()
            finally:
                sys.argv = old_sys_argv

        # Check the C/N logged to stdout
        for line in cm.output:
            if ('C/N' in line):
                cnr_db = float(line.split()[-2])
        self.assertAlmostEqual(cnr_db, 15.95, places=2)



"""
A collection of link budget and other RF calculations.

References:

- [1] L. W. Couch, "Digital & Analog Communication Systems," 8th Ed., 2013.
- [2] M. Lindgren, "A 1296 MHz Earth–Moon–Earth Communication System," 2015.
- [3] T. Pratt and J. E. Allnutt, "Satellite Communications," 3rd Ed., 2019.
- [4] D. M. Pozar, "Microwave Engineering," 4th Ed., 2012.

"""
from math import log10, pi, log2

from . import util
from .constants import T0


def eirp(tx_power, tx_dish_gain):
    """Compute the effective isotropically radiated power (EIRP)

    EIRP (dB) = Tx Power (dB) + Tx Antenna Gain (dB).

    Args:
        tx_power     : Transmit power feeding the antenna (dBW).
        tx_dish_gain : Transmit antenna gain (dB).

    Returns:
        The EIRP in dBW.

    """

    eirp = tx_power + tx_dish_gain
    return eirp


def carrier_eirp(sat_eirp, obo, peb=None, tp_bw=None):
    """Compute the carrier EIRP using transponder/amplifier information

    Converts the saturated EIRP obtained by an amplifier or transponder into
    the corresponding carrier EIRP, after output backoff and power allocation.

    There are two main use cases for this function:

    1. To compute the carrier EIRP from the transponder's saturated EIRP and
       carrier output backoff (OBO).

    2. To compute the carrier EIRP based on the transponder's saturated EIRP,
       the transponder's OBO, the carrier's power-equivalent bandwidth (PEB),
       and the transponder's bandwidth.

    In scenario 1, we start with the transponder's EIRP in dBW obtained when
    its amplifier operates in saturation. This metric is often given as the
    transponder's downlink EIRP at beam peak (BP) or beam center (BC). Then,
    assuming the transponder is shared by multiple FDMA carriers, we subtract
    the carrier OBO to obtain the EIRP assigned to the carrier of interest. In
    this case, note the **carrier OBO** is often a much larger value than the
    **transponder OBO**. The carrier OBO is easily in excess of 10 dB,
    depending on how much of the transponder it occupies. In contrast, as
    discussed in [3], the transponder OBO is typically within the range from 1
    to 7 dB in FDMA systems.

    The OBO interpretation changes in scenario 2 when considering FDMA
    systems. In this case, the OBO is interpreted as the **transponder OBO**,
    not the **carrier OBO**. Hence, the first step in the computation is to
    obtain the transponder's operating EIRP, which is given by the saturated
    EIRP minus the transponder OBO. The second step is to allocate a fraction
    of the available transponder EIRP to the carrier of interest. This fraction
    is determined by the ratio between the PEB and the transponder bandwidth.

    To compute scenario 1, call this function with parameter `obo` as the
    carrier OBO, while leaving the PEB undefined. To compute scenario 2, call
    this function with `obo` as the transponder OBO, while informing the PEB
    and transponder bandwidth. In any case, however, the `sat_eirp` parameter
    represents the transponder's saturated EIRP in the direction of interest.

    Note that, for convenience, the saturated EIRP is **in the direction of
    interest**, not necessarily at BP or BC. For example, it could be at beam
    edge (BE) or any arbitrary coverage contour. Furthermore, not this function
    takes the saturated EIRP (i.e., including the on-axis antenna gain),
    instead of the amplifier's saturated output power. This is intentional
    because it is often easier to find the EIRP information for a given
    satellite contour than to find the transponder and antenna gain separately.

    Lastly, note that both use cases are equivalent in TDMA systems. Since
    there is a single carrier in TDMA, the transponder OBO is the same as the
    carrier OBO. Furthermore, the PEB is the full transponder bandwidth in
    TDMA, so the `peb` parameter does not need to be informed.

    Args:
        sat_eirp  : Saturated amplifier/transponder EIRP in dBW in the
                    direction if interest.
        obo       : Output backoff in dB.
        peb       : Power equivalent bandwidth if the carrier shares the
                    amplifier/transponder with other FDMA carriers.
        tp_bw     : Transponder bandwidth used when the PEB is given.

    Returns:
        Carrier EIRP in dBW.

    """
    if (peb is None):
        carrier_eirp = sat_eirp - obo
    else:
        if (tp_bw is None):
            raise ValueError("The transponder bandwidth is required when "
                             "computing the carrier EIRP with the PEB")
        tp_eirp = sat_eirp - obo
        carrier_eirp = tp_eirp + util.lin_to_db(peb / tp_bw)

    return carrier_eirp


def _path_loss(d, freq):
    """One-way free-space path loss

    Args:
        d    : Distance in meters.
        freq : Carrier frequency in Hz.

    Returns:
        Path loss in dB.

    """
    wavelength = util.wavelength(freq)
    return 20 * log10(4 * pi * d / wavelength)


def path_loss(d, freq, radar=False, obj_gain=None, bistatic=False, d_rx=None):
    """Calculate the free-space path loss (or transmission loss)

    This function supports radar mode, in which case it computes the
    transmission loss considering the path loss in the forward and reverse
    paths (to and from the radar object) as well as the scattering at the radar
    object based on its radar cross section.

    Args:
        d        : Distance in meters between transmitter and receiver (e.g.,
                   satellite to ground station) or between the transmitter and
                   the radar object.
        freq     : Carrier frequency in Hz.
        radar    : Radar mode.
        obj_gain : Gain of the radar object in dB.
        bistatic : Bistatic radar mode.
        d_rx     : Bistatic radar mode only: distance between the radar object
                   and the Rx that is not collocated with the Tx.

    Returns:
        Path loss in dB.

    """
    # Eq. 8-11 from [1], or Eq. 3.16 from [2]:
    Lfs_one_way_db = _path_loss(d, freq)

    if (radar):
        if (obj_gain is None):
            raise ValueError("Radar object's gain is required in radar mode")

        if (bistatic):
            if (d_rx is None):
                raise ValueError("Rx distance required in bistatic radar mode")

            Lfs_tx_db = Lfs_one_way_db
            Lfs_rx_db = _path_loss(d_rx, freq)
            util.log_result("Uplink path loss", "{:.2f} dB".format(Lfs_tx_db))
            util.log_result("Downlink path loss",
                            "{:.2f} dB".format(Lfs_rx_db))
            # Bistatic radar transmission loss in dB, equation 3.24 in [2]:
            Lfs_db = Lfs_tx_db + Lfs_rx_db - obj_gain
        else:
            util.log_result("One-way path loss",
                            "{:.2f} dB".format(Lfs_one_way_db))
            # Monostatic radar transmission loss in dB, equation 3.26 in [2]:
            Lfs_db = 2 * Lfs_one_way_db - obj_gain

        util.log_result("Total path loss", "{:.2f} dB".format(Lfs_db))
    else:
        Lfs_db = Lfs_one_way_db
        util.log_result("Path loss", "{:.2f} dB".format(Lfs_db))
    return Lfs_db


def radar_obj_gain(freq, rcs):
    """Compute the gain of the radar object

    Args:
        freq : Carrier frequency in Hz.
        rcs  : Radar cross section (RCS).

    Returns:
        Gain of the radar object in dB.

    Note:
        The RCS definition repeated in [2] is the following: "the RCS of a
        radar object is the hypothetical area intercepting that amount of power
        which, when scattered isotropically, produces a power density at the
        receiver equal to that from the actual object."

    """
    wavelength = util.wavelength(freq)

    # Radar object gain in dB from equation 3.23 in [2]:
    G_obj_db = 10 * log10(4 * pi * rcs / (wavelength**2))
    util.log_result("Radar object's gain", "{:.2f} dB".format(G_obj_db))
    return G_obj_db


def passive_attn_in_noise_temp(attn_db, T):
    """Input noise temperature of a passive two-port attenuator

    Compute the equivalent input noise temperature of a passive two-port lossy
    component, such as a transmission line or waveguide.

    When considering the equivalent **input** noise temperature, the model is
    as follows:

    .. code-block::

        Signal --> Sum --> Noiseless Passive Component --> Signal + Noise Out
                    ^
                    |
                    |
              Noise Source (Resistor at Temperature Te)

    The noise source is at the input of the passive two-port element, and the
    latter is assumed noiseless but lossy with gain G or attenuation L. The
    noise produced by the hypothetical noise source, when amplified by G (or
    attenuated by L), yields the same noise power on the component's output as
    the noisy (but otherwise equivalent) component produces by itself.

    The output noise power produced by the noisy passive two-port component is
    modeled as equal to :math:`G k T_e B`, namely the input noise power
    :math:`k T_e B` amplified by gain G. Finally, :math:`T_e` is the physical
    temperature of a hypothetical resistor used to represent the noise source.
    The resistor produces noise power :math:`k T_e B` over bandwidth B, and
    :math:`T_e` is called the equivalent input noise temperature, given by:

    .. math::

        T_e = (L - 1) T,

    where :math:`T` is the physical temperature of the two-port component, and
    `L` is the component's attenuation (or loss) in absolute units. See
    Equation (8-31a) in [1] or Equation (10.15) in [4].

    Args:
        attn_db (float): Attenuation in dB.
        T (float): Physical temperature of the attenuator.

    Return:
        float: The equivalent input noise temperature of the passive two-port
        attenuator.

    """
    attn = util.db_to_lin(attn_db)
    return (attn - 1) * T


def passive_attn_out_noise_temp(attn_db, T):
    """Output noise temperature of a passive two-port attenuator

    Compute the equivalent output noise temperature of a passive two-port lossy
    component, such as a transmission line or waveguide.

    When considering the equivalent **output** noise temperature, the model is
    as follows:

    .. code-block::

        Signal --> Noiseless Passive Component --> Sum --> Signal + Noise Out
                                                    ^
                                                    |
                                                    |
                              Noise Source (Resistor at Temperature Teo)

    The noise source is at the output of the passive two-port element, and the
    latter is assumed noiseless but lossy with gain G or attenuation L. The
    hypothetical noise source produces the same noise power as the noisy (but
    otherwise equivalent) two-port component produces by itself.

    The output noise power produced by the noisy passive two-port component is
    modeled as equal to :math:`k T_{eo} B`, and :math:`T_{eo}` is called the
    equivalent output noise temperature. Finally, :math:`T_{eo}` is interpreted
    as the physical temperature of a hypothetical resistor representing the
    noise source, and is given by:

    .. math::

        T_{eo} = (1 - G) T,

    where :math:`T` is the physical temperature of the two-port component, and
    `G` is the component's gain in absolute units (less than unit because the
    component is lossy). See Equation (4.21) in [3].

    Note:
        The equivalent output noise temperature :math:`T_{eo}` is equal to the
        equivalent input noise temperature :math:`T_e` multiplied by the gain G
        of the lossy two-port component or network.

    Args:
        attn_db (float): Attenuation in dB.
        T (float): Physical temperature of the attenuator.

    Return:
        float: The equivalent output noise temperature of the passive two-port
        attenuator.

    """
    gain = 1 / util.db_to_lin(attn_db)
    return (1 - gain) * T


def passive_attn_noise_fig(attn_db, T):
    """Noise figure of a passive two-port attenuator

    Compute the noise figure of a passive two-port lossy component, such as a
    transmission line or waveguide.

    The equivalent input noise temperature of such a component is equal to:

    .. math::

        T_e = (L - 1) T,

    where :math:`T` is the physical temperature of the component, and `L`
    represents its attenuation (or loss) in absolute units.

    Hence, the noise factor referenced to the standard temperature T0=290 K is
    equal to:

    .. math::

        F = 1 + \\frac{T_e}{T0} = 1 + \\frac{T}{T0}(L - 1).

    See Equation 8.32a in [1] or Equation 10.16 in [4].

    Note:
        The noise factor approaches unit if the component's physical
        temperature :math:`T` is close to 290 K. This is typically the case in
        environments inhabitable by humans. In this scenario, the noise figure
        (in dB) is approximately equal to the device's loss in dB. For
        instance, a waveguide or transmission line with 2 dB loss would have a
        noise figure close to 2 dB.

    Args:
        attn_db (float): Attenuation in dB.
        T (float): Physical temperature of the attenuator.

    Returns:
        float: The noise figure of the passive two-port attenuator.

    """
    noise_temp = passive_attn_in_noise_temp(attn_db, T)
    return noise_temp_to_noise_fig(noise_temp)


def antenna_noise_temp(attn_db, T_medium=270, coupling_eff=1.0):
    """Compute the antenna noise temperature

    Follow the theory in Section 4.5.2 of [3].

    Args:

        attn_db      : Total path attenuation (in dB) experienced through the
                       atmosphere, including clear air and rain attenuation.
        T_medium     : Medium temperature (in K) assumed for the rain.
        coupling_eff : Sky noise coupling coefficient determining the fraction
                       of incident sky noise energy output by the antenna.

    Note:
        In [3], a medium temperature of 270 K is considered when computing the
        sky noise temperature for rain. In contrast, a temperature of 290 K is
        considered when analyzing clear sky. The rationale is to be confirmed.

    """
    # Sky's equivalent output noise temperature, assuming the sky behaves as a
    # passive two-port attenuator (see Eq. 4.30 in [3]):
    T_sky = passive_attn_out_noise_temp(attn_db, T_medium)
    # Antenna noise temperature (Eq. 4.31)
    return coupling_eff * T_sky


def coax_loss_nf(length_ft, Tl=T0):
    """Compute the loss and noise figure of a coaxial RG6 transmission line

    Args:
        length_ft : Line length in feet.
        Tl        : temperature of the line in Kelvin.

    Returns:
        Tuple with line loss (dB) and noise figure (dB).

    """
    loss_db_per_ft = 8 / 100
    loss_db = length_ft * loss_db_per_ft
    noise_fig = passive_attn_noise_fig(loss_db, Tl)
    util.log_result("Coax loss", "{:.2f} dB".format(loss_db))
    util.log_result("Coax noise figure", "{:.2f} dB".format(noise_fig))
    return loss_db, noise_fig


def cascaded_noise_figure(nfs, gains):
    """Compute the overall noise figure of cascaded linear devices

    Based on Equation 8-34 from [1].

    Args:
        nfs : List with the noise figures (in dB) corresponding to each
            of the devices in order from input to output.
        gains : List with the gains (also in dB) of the cascaded linear
            devices, excluding the last, in the same order as given for the
            noise figures.

    Note:
        The list of gains should not include the gain of the last device in the
        chain because this gain is irrelevant for the computation.

    Returns:
        The overall noise figure in dB.

    """

    assert (len(nfs) > 0)
    assert (len(gains) == len(nfs) - 1)

    if (len(nfs) == 1):
        return nfs[0]

    F = util.db_to_lin(nfs[0])
    G_prod = 1
    for i, nf in enumerate(nfs[1:]):
        nf_abs = util.db_to_lin(nf)
        G_prod *= util.db_to_lin(gains[i])
        F += (nf_abs - 1) / G_prod

    F_db = 10 * log10(F)
    util.log_result("Rx noise figure", "{:.2f} dB".format(F_db))
    return F_db


def cascaded_input_noise_temp(temps, gains):
    """Equivalent input noise temperature of cascaded linear devices

    Computes the overall equivalent input noise temperature for cascaded linear
    devices according to Equation 8-37 from [1].

    Args:
        temps : List with the individual input noise temperatures (in K)
            corresponding to each of the devices in order from input to output.
        gains : List with the gains (in dB) of the cascaded linear devices,
            excluding the last, in the same order as given for the input noise
            temperatures.

    Note:
        The list of gains should not include the gain of the last device in the
        chain because this gain is irrelevant for the computation.

    Returns:
        The overall input noise temperature in K.

    """
    assert (len(temps) > 0)
    assert (len(gains) == len(temps) - 1)

    if (len(temps) == 1):
        return temps[0]

    Te = temps[0]
    G_prod = 1
    for i, temp in enumerate(temps[1:]):
        G_prod *= util.db_to_lin(gains[i])
        Te += temp / G_prod

    return Te


def noise_fig_to_noise_temp(nf):
    """Convert noise figure to the effective input-noise temperature

    Note the noise figure is always referenced to a noise source at the
    standard noise temperature of T0 = 290 K. In contrast, the noise
    temperature is independent of the temperature of the noise source.

    Args:
        nf : Noise figure in dB.

    Returns:
        Noise temperature in K.

    """
    nf_abs = util.db_to_lin(nf)

    # Using Equation 8-30b in [1]:
    Te = T0 * (nf_abs - 1)

    return Te


def noise_temp_to_noise_fig(Te):
    """Convert an effective input-noise temperature to a noise figure in dB

    Args:
        Te : Noise temperature in K.

    Returns:
        Noise figure in dB.

    """
    # Noise factor
    nf_abs = 1 + Te / T0
    # Return the noise figure
    return 10 * log10(nf_abs)


def rx_sys_noise_temp(Tar, Te, feed_loss_db=0, feed_temp=T0):
    """Compute the receiver system noise temperature

    The receiver system noise temperature is equal to the effective input noise
    temperature (Te) of the entire receiver (seen as a blackbox) added to the
    antenna noise temperature (Tar). The Te term represents the noise
    introduced by the cascaded linear components of the receiver (e.g., LNB,
    coax line, and radio interface). Meanwhile, The Tar component represents
    the cosmic noise and Earth blackbody radiation captured by the antenna.
    Hence, the simplified model is as follows:

    .. code-block::

         Rx Antenna (Tar) -----> Sum ----> Noise-free Gain Stage ---> Detector
                                  ^
                                  |
                                  |
                           Receiver Noise (Te)

    Note the antenna is not treated as another cascaded device within the
    receiver. Instead, the antenna noise adds to the equivalent input noise of
    the cascaded devices, see Figure 8-24 in [1]. This is because the
    observation point is at the input to the receiver system. As explained in
    [2], around equation 4.39, this is a convenient choice in terms of where
    the effective input noise temperature is observed.

    By definition, an equivalent (or effective) input noise temperature
    represents the thermodynamic temperature of a noisy resistor connected to
    the input of a noiseless two-port element, which gives the same output
    noise power as the noisy (but otherwise equivalent) two-port element with
    an ideal noiseless source at its input [2]. Hence, if we group the entire
    receiver into a single equivalent two-port element, the Te term represents
    the noise generated by the entire receiver, which has power k*Te*B. When
    combined to the noise collected by the antenna, the result becomes the
    total system noise temperature.

    Since the observation point is at the Rx input, the carrier-to-noise ratio
    (CNR) becomes the carrier power captured by the antenna (i.e., at the Rx
    input) divided by k*Te*B. If the observation point were instead at the
    detector's input, both the carrier and noise powers would be multiplied by
    the combined gain of the cascaded linear components. Hence, the CNR would
    remain the same (see Equation 4.15 in [3]), which confirms the convenience
    of using the receiver's input as the reference point.

    Besides, note the input to the receiver typically represents the input to
    the LNA, which is assumed to be directly connected to the antenna. However,
    in practice, there is a lossy passive two-port element (waveguide or
    transmission line) connecting the antenna to the LNA. When the loss in this
    segment is non-negligible, the model becomes:

    .. code-block::

        Antenna (Tar) --> Waveguide/Line --> Sum --> Noiseless Rx --> Detector
                                              ^
                                              |
                                              |
                                       Receiver Noise (Te)

    In this case, the lossy waveguide or transmission line attenuates the
    antenna noise, but also generates noise of its own. This function takes
    both effects into account when a non-zero feed loss is specified.

    Args:
        Tar (float): Antenna noise temperature in K.
        Te (float): Effective input-noise temperature in K.
        feed_loss_db (float): Loss (in dB) over the passive two-port feed
            connecting the antenna to the LNA.
        feed_temp (float): Physical temperature of the feed.

    Returns:
        Receiver system noise temperature in K.

    """

    util.log_result("Antenna noise temperature", "{:.2f} K".format(Tar))

    if (feed_loss_db > 0):
        # Equivalent output noise temperature of the feed
        T_feed = passive_attn_out_noise_temp(feed_loss_db, feed_temp)
        # Antenna noise temperature at the feed's output, i.e., after
        # attenuation (see Example 4.4 in [3]):
        attn = util.db_to_lin(feed_loss_db)
        Tar = Tar / attn
        # Total noise (antenna + feed) into the receiver:
        Tin = Tar + T_feed
        util.log_result("Lossy feed noise temperature",
                        "{:.2f} K".format(T_feed))
        util.log_result("Antenna/feed noise temperature",
                        "{:.2f} K".format(Tin))
    else:
        Tin = Tar

    # Equation 8-41 from [1], or 4.39 from [2]:
    Tsyst = Tin + Te
    util.log_result(
        "System noise temperature",
        "{:.2f} K ({:.2f} dBK)".format(Tsyst, util.lin_to_db(Tsyst)))
    return Tsyst


def g_over_t(rx_ant_gain_db, T_sys_db):
    """Compute the G/T ratio in dB/K

    Compute the ratio between the Rx antenna gain and the receiver system noise
    temperature, known as G/T. This ratio determines the quality of the
    satellite receiving system. The CNR is proportional to it, and the G/T
    ratio represents the part of the CNR that can be improved based on the
    receiver hardware alone. The other terms of the CNR are EIRP, slant range,
    noise bandwidth, and downlink frequency, which are usually fixed for a
    specific location and service.

    Args:
        rx_ant_gain_db : Receiver antenna gain in dB.
        T_sys_db       : Receiver system noise temperature in dBK.

    Returns:
        G/T in decibels with units of dBK^−1 (or dB/K).

    """
    g_over_t_db = rx_ant_gain_db - T_sys_db
    util.log_result("G/T", "{:.2f} dB/K".format(g_over_t_db))
    return g_over_t_db


def rx_flux_density(eirp_db, distance, atm_loss_db=0):
    """Compute the received flux density in in dBW/m^2

    Note the Rx flux density refers the incident power density (power per area)
    arriving at the receiving antenna, i.e., at the antenna's input. In
    contrast, the value computed by function "rx_power" refers to the power
    collected by the antenna, i.e., at the antenna's output.

    Args:
        eirp_db     : EIRP in dBW.
        distance    : Distance between the Tx and Rx in meters.
        atm_loss_db : Total atmospheric loss in dB.

    Returns:
        Received flux density in dBW/m^2.

    Note:

        The flux density unit of :math:`\\text{dBW}/m^2` should be interpreted
        as :math:`10\\log10(\\frac{W}{m^2})`, and not
        :math:`\\frac{10\\log10(W)}{m^2}`. With a flux density F in
        :math:`\\text{dBW}/m^2` and a given area A in :math:`m^2`, the Rx power
        can be obtained by :math:`F + 10\\log10(A)`. Equivalently, if the area
        is given as :math:`A_{db}` in :math:`\\text{dBm}^2` (decibels greater
        than 1 square meter), the Rx power can be obtained :math:`F + A_{db}`.

    """
    Pt = eirp_db - atm_loss_db  # Tx power minus atmospheric losses
    flux = util.db_to_lin(Pt) / (4 * pi * distance**2)  # in W / m^2
    flux_dbw_m2 = util.lin_to_db(flux)
    util.log_result("Rx flux density", "{:.2f} dBW/m2".format(flux_dbw_m2))
    return flux_dbw_m2


def rx_power(eirp_db,
             path_loss_db,
             rx_ant_gain_db,
             atm_loss_db=0,
             mispointing_db=0,
             feed_loss_db=0):
    """Compute the received carrier power in dBW

    Args:
        eirp_db        : EIRP in dBW.
        path_loss_db   : Free-space path loss in dB.
        rx_ant_gain_db : Receiver antenna gain in dB.
        atm_loss_db    : Total atmospheric loss in dB.
        mispointing_db : Antenna mispointing loss in dB.
        feed_loss_db   : Antenna-to-LNA feed loss in dB.

    Returns:
        Received power in dBW.

    """
    P_rx_dbw = eirp_db - path_loss_db - atm_loss_db + rx_ant_gain_db \
        - mispointing_db - feed_loss_db

    util.log_result("Rx Power (LNB Input)",
                    "{:.2f} dBm".format(util.dbw_to_dbm(P_rx_dbw)))

    return P_rx_dbw


def noise_power(T_sys_db, bw):
    """Compute the receiver noise power in dBW

    According to Equation 8-40 in [1], the noise power is given by `N =
    k*Tsyst*bw`, where k is the Boltzmann constant, Tsyst is the receiver
    system noise temperature (in absolute units) and bw is the IF equivalent
    bandwidth in Hz.

    Args:
        T_sys_db : Receiver system noise temperature in dBK.
        bw       : Nominal signal bandwidth (also known as noise bandwidth).

    Returns:
        Receiver noise power in dBW.

    """

    k_db = -228.6  # Boltzmann’s constant (of 1.38e-23) in dB
    bw_db = 10 * log10(bw)

    # Noise power:
    N_dbw = k_db + T_sys_db + bw_db
    util.log_result("Noise Power", "{:.2f} dBm".format(util.dbw_to_dbm(N_dbw)))

    return N_dbw


def spectral_density(power_dbw, bw, label=""):
    """Compute the power spectral density (PSD) of a given signal

    Assume the signal has a constant power spectral density over the given
    bandwidth. Equivalently, assume the signal behaves as white noise, i.e.,
    with uncorrelated zero-mean samples in time.

    Args:
        power_dbw : Signal power in dBW.
        bw        : Bandwidth in Hz.
        label     : Optional label used for logging.

    Returns:
        PSD in dBW/Hz.

    """
    bw_db = util.lin_to_db(bw)
    psd_dbw_hz = power_dbw - bw_db
    util.log_result("{} PSD".format(label),
                    "{:.2f} dBm/Hz".format(util.dbw_to_dbm(psd_dbw_hz)))
    return psd_dbw_hz


def cnr(P_rx_dbw, N_dbw):
    """Compute the carrier-to-noise ratio (CNR) in dB

    Args:
        P_rx_dbw : Received (carrier) power in dBW
        N_dbw    : Receiver noise power in dBW

    Returns:
        CNR in dB.

    """
    cnr_db = P_rx_dbw - N_dbw
    util.log_result("C/N", "{:.2f} dB".format(cnr_db))

    return cnr_db


def capacity(snr_db, bw):
    """Compute the channel capacity in bps

    Args:
        snr_db : signal-to-noise ratio in dB.
        bw     : nominal bandwidth.

    Returns:
        Capacity in bits per second (bps).

    """
    snr = util.db_to_lin(snr_db)
    c = bw * log2(1 + snr)
    util.log_result("Capacity", util.format_rate(c))
    return c


def carrier_to_asi_ratio(rx_dish, long_separation, asi_eirp_ratio):
    """Compute the carrier to adjacent satellite interference (ASI) ratio

    Args:
        rx_dish (Antenna): Rx antenna object.
        long_separation (float): Longitudinal separation in degrees between
            the wanted satellite and the adjacent interferer(s).
        asi_eirp_ratio (float): Ratio between the aggregate adjacent downlink
            EIRP and the wanted signal's EIRP.

    Returns:
        (float): Carrier-to-ASI ratio in dB.

    """
    # Compute the difference between the on-axis (boresight) antenna gain and
    # the off-axis gain at the specified longitudinal separation.
    on_axis_gain = rx_dish.gain_db
    off_axis_gain = rx_dish.off_axis_gain(long_separation)
    delta_gain_db = on_axis_gain - off_axis_gain
    # The adjacent signal is received with the off-axis gain. Hence, the
    # difference between the wanted carrier and the adjacent signal power is
    # delta_gain_db if the adjacent signal has the same EIRP as the wanted
    # signal. Otherwise, if the aggregate adjacent downlink EIRP is greater
    # than the EIRP from the wanted carrier (asi_eirp_ratio > 1), the
    # carrier-to-asi ratio reduces. Similarly, in the opposite case, when
    # asi_eirp_ratio < 1, the carrier-to-asi increases.
    c_to_i_db = delta_gain_db - util.lin_to_db(asi_eirp_ratio)
    util.log_result("ASI C/I ({}° separation)".format(long_separation),
                    "{:.2f} dB".format(c_to_i_db))
    return c_to_i_db


def cnir(cnr, ci):
    """Compute the carrier to noise plus interference ratio

    Based on the reciprocal CNR formula in Eq. (4.42) from [3].

    Args:
        cnr (float): Carrier-to-noise ratio in dB.
        ci (float): Carrier-to-interference ratio in dB.

    Returns:
        (float): Carrier to noise plus interference ratio C/(N+I) in dB.

    """
    cnr_lin = util.db_to_lin(cnr)
    ci_lin = util.db_to_lin(ci)
    cnir_lin = 1 / ((1 / cnr_lin) + (1 / ci_lin))
    cnir_db = util.lin_to_db(cnir_lin)
    util.log_result("C/(N+I)", "{:.2f} dB".format(cnir_db))
    return cnir_db



"""
References:

 [1] Couch, Leon W.. Digital & Analog Communication Systems.
 [2] www.everythingrf.com/rf-calculators/parabolic-reflector-antenna-gain

"""

import unittest

import numpy as np

from . import antenna


class TestAntenna(unittest.TestCase):

    def test_dish_gain(self):
        """Test dish gain calculation"""
        # Example 8-5 in [1]:
        dish = antenna.Antenna(freq=4e9, diameter=3.05, efficiency=0.557)
        self.assertAlmostEqual(
            dish.gain_db,
            39.6,  # expected gain in dB
            places=1)

        # Using [2] with an aperture efficiency of 56%:
        dish = antenna.Antenna(freq=12.45e9, diameter=0.45, efficiency=0.56)
        self.assertAlmostEqual(
            dish.gain_db,
            32.84568544,  # expected gain in dB
            places=1)

    def test_dish_aperture_inference(self):
        """Test inference of the effective aperture area"""
        freq = 4e9  # antenna's operating frequency

        # Create two antenna objects. The first can calculate the effective
        # aperture area directly based on the given physical area and the
        # aperture efficiency. The second is created directly with the gain
        # from dish 1, in which case the effective aperture area is inferred
        # only. The two effective aperture computations should match.
        dish1 = antenna.Antenna(freq=freq, diameter=3.05, efficiency=0.557)
        dish2 = antenna.Antenna(freq=freq, gain=dish1.gain_db)
        self.assertAlmostEqual(dish1.effective_aperture,
                               dish2.effective_aperture)

    def test_dish_diameter_inference(self):
        """Test inference of the physical diameter"""
        freq = 12.45e9
        # Dish object with diameter informed by argument
        dish1 = antenna.Antenna(freq, diameter=0.45, efficiency=0.56)
        # Dish object informing the gain and the aperture efficiency only
        dish2 = antenna.Antenna(freq, gain=dish1.gain_db, efficiency=0.56)
        # Verify that the diameter can be inferred correctly
        self.assertEqual(dish1.diameter, dish2.diameter)
        # If the efficiency is not provided, the physical diameter cannot be
        # inferred. In this case, a warning is expected:
        with self.assertLogs(level='WARNING'):
            dish3 = antenna.Antenna(freq, gain=dish1.gain_db)
            self.assertIsNone(dish3.diameter)

    def test_required_params(self):
        """Test the required constructor parameters"""

        # The operating frequency is always required:
        with self.assertRaises(TypeError):
            antenna.Antenna(diameter=3.05, efficiency=0.557)

        with self.assertRaises(ValueError):
            antenna.Antenna(freq=None, diameter=3.05, efficiency=0.557)

        with self.assertRaises(ValueError):
            antenna.Antenna(freq=None, gain=30)

        # When the gain is not provided, both the diameter and aperture
        # efficiency are required.
        with self.assertRaises(ValueError):
            antenna.Antenna(freq=4e9, efficiency=0.557)

        with self.assertRaises(ValueError):
            antenna.Antenna(freq=4e9, diameter=3.05)

    def test_antenna_off_axis_gain(self):
        """Test off-axis gain computation

        Test a couple of arbitrary values, most of which can be computed
        directly as "32 - 25*log10(angle)".

        """
        # Antenna with diameter/wavelength = 23.4:
        dish1 = antenna.Antenna(freq=11.7e9, diameter=0.6, efficiency=0.65)
        self.assertAlmostEqual(dish1.gain_db, 35.5, places=1)
        self.assertAlmostEqual(dish1.off_axis_gain(0), dish1.gain_db, places=1)
        self.assertAlmostEqual(dish1.off_axis_gain(0.1), 35.4, places=1)
        self.assertAlmostEqual(dish1.off_axis_gain(1), 34.1, places=1)
        self.assertAlmostEqual(dish1.off_axis_gain(2), 30, places=1)
        self.assertAlmostEqual(dish1.off_axis_gain(3), 20.1, places=1)
        self.assertAlmostEqual(dish1.off_axis_gain(4), 16.9, places=1)
        self.assertAlmostEqual(dish1.off_axis_gain(5), 14.5, places=1)
        self.assertAlmostEqual(dish1.off_axis_gain(10), 7, places=1)
        self.assertAlmostEqual(dish1.off_axis_gain(20), -0.5, places=1)
        self.assertAlmostEqual(dish1.off_axis_gain(30), -4.92, places=1)
        self.assertAlmostEqual(dish1.off_axis_gain(80), -10, places=1)

        # Antenna with diameter/wavelength = 46.8 (phi_1 = 1.55):
        dish2 = antenna.Antenna(freq=11.7e9, diameter=1.2, efficiency=0.65)
        self.assertAlmostEqual(dish2.gain_db, 41.5, places=1)
        # The gains within the main lobe differ from the 0.6m antenna
        self.assertAlmostEqual(dish2.off_axis_gain(0), dish2.gain_db, places=1)
        self.assertAlmostEqual(dish2.off_axis_gain(0.1), 41.4, places=1)
        self.assertAlmostEqual(dish2.off_axis_gain(1), 36, places=1)
        self.assertAlmostEqual(dish2.off_axis_gain(2), 24.47, places=1)
        # In contrast, the sidelobe values follow the same expression:
        self.assertAlmostEqual(dish2.off_axis_gain(3), 20.1, places=1)
        self.assertAlmostEqual(dish2.off_axis_gain(4), 16.9, places=1)
        self.assertAlmostEqual(dish2.off_axis_gain(5), 14.5, places=1)
        self.assertAlmostEqual(dish2.off_axis_gain(10), 7, places=1)
        self.assertAlmostEqual(dish2.off_axis_gain(20), -0.5, places=1)
        self.assertAlmostEqual(dish2.off_axis_gain(30), -4.92, places=1)
        self.assertAlmostEqual(dish2.off_axis_gain(80), -10, places=1)

        # Ensure the antenna gain is monotonically decreasing
        for dish in [dish1, dish2]:
            last_gain = np.inf
            for angle in np.arange(0, 180, 1e-2):
                gain = dish.off_axis_gain(angle)
                self.assertGreaterEqual(last_gain, gain,
                                        "Failed for angle {}".format(angle))
                last_gain = gain

        # The off-axis angle must be >=0 and <= 180°:
        with self.assertRaises(ValueError):
            dish.off_axis_gain(-1)

        with self.assertRaises(ValueError):
            dish.off_axis_gain(180.01)

        # The model is valid for frequencies within [2, 31] GHz:
        freq = 1.9e9
        dish = antenna.Antenna(freq=freq, diameter=0.45, efficiency=0.65)
        with self.assertRaises(ValueError):
            dish.off_axis_gain(0)

        freq = 32.1e9
        dish = antenna.Antenna(freq=freq, diameter=0.45, efficiency=0.65)
        with self.assertRaises(ValueError):
            dish.off_axis_gain(0)



"""Link budget command-line utility"""
import argparse
import json
import logging
import sys

from dateutil.parser import isoparse

from . import calc, constants, propagation, pointing, util
from .antenna import Antenna

__version__ = "0.1.10"


def get_parser():
    """Command-line arguments"""
    parser = argparse.ArgumentParser(
        description="Link Budget Calculator",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    general_p = parser.add_argument_group('General Options')
    general_p.add_argument('--json',
                           action='store_true',
                           help='Print results in JSON format.')
    general_p.add_argument('-v',
                           '--version',
                           action='version',
                           version='%(prog)s {}'.format(__version__))

    margin_p = parser.add_argument_group('Margin Options')
    margin_p.add_argument(
        '--min-cnr',
        type=float,
        help='Target minimum carrier-to-noise ratio (CNR) in dB for the '
        'link margin computation.')
    margin_p.add_argument(
        '--impl-margin',
        type=float,
        default=0,
        help='Implementation margin in dB accounting for the non-ideal '
        'behavior of the receiver system.')

    freq_p = parser.add_argument_group('Frequency Options')
    freq_p.add_argument(
        '--freq',
        required=True,
        type=float,
        help='Downlink carrier frequency in Hz for satellite signals or '
        'simply the signal frequency in Hz for radar (passively reflected) '
        'signals.')
    freq_p.add_argument('--bw',
                        '--if-bw',
                        required=True,
                        type=float,
                        help='Nominal signal bandwidth in Hz.')

    pol_p = parser.add_argument_group('Polarization Options')
    pol_p.add_argument(
        '--polarization',
        choices=['circular', 'linear'],
        default='linear',
        help='Polarization of the transmitted electromagnetic wave.')

    tx_feed_p = parser.add_argument_group('Tx Power Options')
    tx_pwr_group = tx_feed_p.add_mutually_exclusive_group(required=True)
    tx_pwr_group.add_argument(
        '--eirp',
        type=float,
        help="Carrier or transponder EIRP in dBW. If an output backoff is "
        "defined, this EIRP corresponds to the amplifier\'s saturated "
        "operation. Otherwise, it represents the actual operating EIRP.")
    tx_pwr_group.add_argument(
        '--tx-power',
        type=float,
        help='Power feeding the Tx antenna in dBW. If an output backoff is '
        'defined, this parameter represents the amplifier\'s saturated '
        'output power. Otherwise, it refers to the actual Tx power. ')
    tx_feed_p.add_argument('--obo',
                           type=float,
                           default=0,
                           help="Carrier or transponder output backoff in dB.")

    fdma_group = parser.add_argument_group(
        title='FDMA Carrier Power Options',
        description="Parameters to determine the power allocated to an FDMA "
        "carrier.")
    fdma_group.add_argument(
        '--carrier-peb',
        type=float,
        help="Power-equivalent bandwidth (PEB) in Hz assigned for the FDMA "
        "carrier. When provided, the EIRP computed from --eirp or --tx-power "
        "refers to the transponder, while the PEB determines the fraction of "
        "this transponder EIRP allocated to the carrier. In this case, note "
        "the output backoff must refer to the transponder, not the carrier. "
        "If the output backoff represents the carrier backoff, do not inform "
        "the carrier PEB.")
    fdma_group.add_argument(
        '--tp-bw',
        type=float,
        help="Transponder bandwidth in Hz, required if the PEB is provided.")

    dish_p = parser.add_argument_group('Antenna Options')
    tx_dish_group = dish_p.add_mutually_exclusive_group()
    tx_dish_group.add_argument(
        '--tx-dish-size',
        type=float,
        help='Diameter in meters of the parabolic antenna used for '
        'transmission. Used when the power is specified through option '
        '--tx-power.')
    tx_dish_group.add_argument(
        '--tx-dish-gain',
        type=float,
        help='Gain in dBi of the parabolic antenna used for transmission. '
        'Used when the power is specified through option --tx-power.')
    dish_p.add_argument(
        '--tx-dish-efficiency',
        type=float,
        default=0.56,
        help='Aperture efficiency of the parabolic antenna used for '
        'transmission. Determines the antenna gain when the dish is specified '
        'by size (option ``--tx-dish-size``). Otherwise, when the gain is '
        'defined directly by option ``--tx-dish-gain``, this parameter is '
        'used to infer the diameter of an equivalent parabolic reflector.')
    rx_dish_group = dish_p.add_mutually_exclusive_group(required=True)
    rx_dish_group.add_argument(
        '--rx-dish-size',
        type=float,
        help='Diameter in meters of the parabolic antenna used for '
        'reception.')
    rx_dish_group.add_argument(
        '--rx-dish-gain',
        type=float,
        help='Gain in dBi of the parabolic antenna used for reception.')
    dish_p.add_argument(
        '--rx-dish-efficiency',
        type=float,
        default=0.56,
        help='Aperture efficiency of the parabolic antenna used for '
        'reception. Determines the antenna gain when the dish is specified by '
        'size (option ``--rx-dish-size``). Otherwise, when the gain is '
        'defined directly by option ``--rx-dish-gain``, this parameter is '
        'used to infer the diameter of an equivalent parabolic reflector.')

    prop_p = parser.add_argument_group('Propagation Options')
    prop_p.add_argument(
        '--atmospheric-loss',
        type=float,
        help='Attenuation in dB experienced through the atmosphere. It should '
        'always include the clear air attenuation, and it could include other '
        'effects such as rain and cloud attenuation. When analyzing a radar '
        'system, note this option should determine the one-way attenuation, '
        'not the two-way. When omitted, the program assumes a reasonable '
        'atmospheric attenuation based on models from ITU-R recommendations.')
    prop_p.add_argument(
        '--availability',
        type=float,
        default=99,
        help='Target link availability in %% to consider on the atmospheric '
        'attenuation model. For instance, when targeting at a 99.9%% '
        'availability, the analysis is based on the atmospheric attenuation '
        'exceeded 0.1%% of the time.')

    interf_p = parser.add_argument_group('Interference Options')
    interf_p.add_argument(
        '--asi',
        action='store_true',
        default=False,
        help='Include adjacent satellite interference (ASI) in the link  '
        'budget considering neighbor satellites with overlapping coverage, '
        'frequency and polarization.')
    interf_p.add_argument(
        '--asi-eirp-ratio',
        type=float,
        default=1.0,
        help='Ratio between the aggregate downlink EIRP from adjacent '
        'satellites and the wanted signal\'s EIRP.')
    interf_p.add_argument(
        '--asi-long-separation',
        type=float,
        default=2.0,
        help='Longitudinal orbit separation in degrees between the wanted '
        'satellite and the adjacent satellite(s).')

    rx_p = parser.add_argument_group('Rx Options')
    rx_p.add_argument(
        '--antenna-noise-temp',
        type=float,
        help='Receive antenna\'s noise temperature in K. When omitted, '
        'this parameter is derived based on the atmospheric attenuation.')
    rx_p.add_argument('--mispointing-loss',
                      type=float,
                      default=0,
                      help='Loss in dB due to antenna mispointing.')
    lnb_noise_group = rx_p.add_mutually_exclusive_group(required=True)
    lnb_noise_group.add_argument('--lnb-noise-fig',
                                 type=float,
                                 help='LNB\'s noise figure in dB.')
    lnb_noise_group.add_argument('--lnb-noise-temp',
                                 type=float,
                                 help='LNB\'s noise temperature in K.')
    rx_p.add_argument('--lnb-gain',
                      required=True,
                      type=float,
                      help='LNB\'s gain.')
    rx_p.add_argument('--rx-noise-fig',
                      required=True,
                      type=float,
                      help='Receiver\'s noise figure in dB.')
    rx_p.add_argument(
        '--coax-length',
        required=True,
        type=float,
        help='Length of the coaxial transmission line between the LNB and the '
        'receiver in ft.')
    rx_p.add_argument(
        '--lna-feed-loss',
        type=float,
        default=0,
        help='Loss in dB of the passive feed (typically waveguide) connecting '
        'the antenna to the LNA. Applicable when a discrete LNB is used '
        'instead of an integrated LNBF.')

    pos_p = parser.add_argument_group(
        title='Earth Station Position Information',
        description="Note longitudes are positive east of the prime meridian "
        "and negative otherwise (west), while latitudes are positive above "
        "the equator (north) and negative otherwise (south).")
    pos_p.add_argument('--rx-long',
                       type=float,
                       help='Rx station\'s longitude in degrees.')
    pos_p.add_argument('--rx-lat',
                       type=float,
                       help='Rx station\'s latitude in degrees.')
    pos_p.add_argument(
        '--rx-height',
        type=float,
        default=0,
        help='Rx station\'s height in meters above mean sea level.')

    pos_s = parser.add_argument_group(
        title='Satellite or Radar Object Position Information',
        description="Note longitudes are positive east of the prime meridian "
        "and negative otherwise (west), while latitudes are positive above "
        "the equator (north) and negative otherwise (south).")
    pos_s.add_argument(
        '--sat-long',
        type=float,
        help='Longitude in degrees of the satellite or radar object.')
    pos_s.add_argument(
        '--sat-lat',
        type=float,
        default=0,
        help='Latitude in degrees of the satellite or radar object.')
    pos_s.add_argument(
        '--sat-alt',
        type=float,
        default=constants.GEOSYNC_ORBIT,
        help='Satellite or radar object\'s altitude in meters above '
        'the reference ellipsoid.')
    pos_s.add_argument(
        '--slant-range',
        type=float,
        help='Slant path length in km between the Rx station and the '
        'satellite or reflector.')
    pos_s.add_argument(
        '--ref-ellipsoid',
        choices=['WGS84', 'GRS80'],
        default='WGS84',
        help="Reference Earth ellipsoid to use in the computation.")
    pos_s.add_argument(
        '--obs-time',
        type=isoparse,
        help="Observation time for the satellite position\'s prediction. "
        "Requires the \'--sat-tle-name\' option and must be given in "
        "ISO 8601 format.")

    tle_p = parser.add_argument_group(title='Satellite TLE Information')
    tle_p.add_argument('--sat-tle-name',
                       help="Satellite name on CelesTrak\'s TLE database.")
    tle_p.add_argument(
        '--sat-tle-group',
        choices=pointing.CELESTRAK_GROUPS,
        default=None,
        help="CelesTrak satellite group to restrict the search. "
        "When undefined, searches the entire database.")
    tle_p.add_argument(
        '--tle-save-dir',
        default=pointing.get_default_tle_dataset_dir(),
        help="Directory where downloaded TLE datasets should be saved.")
    tle_p.add_argument('--tle-no-save',
                       action='store_true',
                       default=False,
                       help="Do not save the downloaded TLE datasets locally. "
                       "A new download will be required every time.")

    radar_p = parser.add_argument_group('Radar Options')
    radar_p.add_argument(
        '--radar',
        default=False,
        action='store_true',
        help='Activate monostatic radar mode so that the link budget '
        'considers the path loss to and back from object.')
    radar_p.add_argument('--radar-cross-section',
                         type=float,
                         help='Radar cross-section of the radar object.')
    return parser


def validate(parser, args):
    """Validate command-line arguments"""

    # Validate the latitude and longitude coordinates
    lat = args.rx_lat
    lng1 = args.rx_long
    lng2 = args.sat_long
    if (lat is not None and (lat < -90 or lat > 90)):
        parser.error("--rx-lat must be within -90 (South) to +90 (North).")

    if (lng1 is not None and (lng1 < -180 or lng1 > 180)):
        parser.error("--rx-long must be within -180 (West) to +180 (East).")

    if (lng2 is not None and (lng2 < -180 or lng2 > 180)):
        parser.error("--sat-long must be within -180 (West) to +180 (East).")

    if (args.tx_power and args.tx_dish_size is None
            and args.tx_dish_gain is None):
        parser.error("Define either --tx-dish-size or --tx-dish-gain  "
                     "using option --tx-power")

    # Link availability must be within [95, 99.999] because that's the range
    # supported by the ITU-R models (see, e.g., Step 10 in Section 2.2.1.1 of
    # ITU-R P.618-13).
    if (args.availability > 99.999 or args.availability < 95):
        parser.error(
            "Target link availability must be between 95 and 99.999 %%")

    if (args.radar):
        if (args.radar_cross_section is None):
            parser.error("Argument --radar-cross-section is required in radar "
                         "mode (--radar)")

    # When the slant range is provided, none of the coordinates matter (neither
    # the satellite's nor the Rx station's). That is because the look angles
    # are not computed when the slant range is given. The slant range suffices
    # for the link budget calculations in general, except if the atmospheric
    # loss models are active, in which case the coordinates are required.
    #
    # Furthermore, when the satellite TLE name is given, the satellite position
    # is obtained by prediction with the TLE orbital data, so only the Rx
    # coordinates are required. However, the user may also choose to inform the
    # satellite coordinates directly.
    if args.slant_range is None:
        if args.sat_tle_name is None:  # satellite position by coordinates
            required_args = {
                '--sat-long': args.sat_long,
                '--rx-long': args.rx_long,
                '--rx-lat': args.rx_lat
            }
            # NOTE the --sat-lat, --sat-alt, and --rx-height args have default
            # values.
            mutex_source = '--sat-long'
            mutex_args = {
                '--sat-tle-name': args.sat_tle_name,
                '--obs-time': args.obs_time
            }
        else:  # satellite position via TLE prediction
            required_args = {
                '--rx-long': args.rx_long,
                '--rx-lat': args.rx_lat
            }
            mutex_source = '--sat-tle-name'
            mutex_args = {'--sat-long': args.sat_long}

    else:  # slant range is given
        required_args = {}
        mutex_source = '--slant-range'
        mutex_args = {
            '--sat-tle-name': args.sat_tle_name,
            '--sat-long': args.sat_long,
            '--rx-long': args.rx_long,
            '--rx-lat': args.rx_lat,
            '--obs-time': args.obs_time
        }

    missing_args = [key for key, val in required_args.items() if val is None]
    if missing_args:
        parser.error("the following arguments are required: {}".format(
            ", ".join(missing_args)))
    unexpected_args = [
        key for key, val in mutex_args.items() if val is not None
    ]
    if unexpected_args:
        parser.error("argument{} {}: not allowed with argument {}".format(
            "s" if len(unexpected_args) > 1 else "",
            ", ".join(unexpected_args), mutex_source))

    # The atmospheric loss models require the Rx/satellite coordinates.
    if args.slant_range is not None and args.atmospheric_loss is None:
        logging.error(
            "Unable to model the atmospheric loss without the Rx and "
            "satellite coordinates")
        parser.error(
            "Define the --atmospheric-loss or substitute --slant-range by the "
            "Rx/satellite coordinates")

    # If the antenna noise temperature is given directly, it will take
    # precedence over the inferred temperature based on the atmospheric
    # loss. This approach makes more sense when the atmospheric loss is also
    # given directly. When the atmospheric loss is not given directly, it is
    # determined automatically from ITU-R models. In this case, it makes sense
    # to let the antenna noise temperature be determined by the same models
    # too. Warn the user to prevent unintenional usage of the option.
    if (args.atmospheric_loss is None and args.antenna_noise_temp is not None):
        logging.warning(
            "Atmospheric loss model used to determine the atmospheric "
            "attenuation but not the antenna noise temperature (overriden by "
            "option --antenna-noise-temp)")

    if (args.carrier_peb is not None and args.tp_bw is None):
        parser.error("Argument --tp-bw is required if option --carrier-peb "
                     "is defined")

    # Validate the implementation margin
    if (args.impl_margin < 0):
        parser.error("argument --impl-margin must be a non-negative number")


def configure_logging():
    """Configure the logging format"""
    logging_fmt = "%(levelname)s %(message)s"
    logging.basicConfig(level=logging.INFO, format=logging_fmt)
    # Keep the level name except if it's an INFO message
    logging.addLevelName(logging.INFO, '')


def analyze(args, verbose=False):
    """Main link budget analysis

    Args:
        args : Populated argparse namespace object.
        verbose : Verbose mode.

    Returns:
        Dictionary with the main link budget results.

    """
    if (verbose and not args.json):
        util.log_header()

    # -------- Look angles --------
    if (args.slant_range is None):
        if args.sat_tle_name is not None:
            try:
                sat_long, sat_lat, sat_alt = pointing.get_sat_pos_by_tle(
                    args.sat_tle_name, args.sat_tle_group, args.obs_time,
                    args.tle_save_dir, args.tle_no_save)
            except ValueError:
                sys.exit(1)
        else:
            sat_long = args.sat_long
            sat_lat = args.sat_lat
            sat_alt = args.sat_alt

        # Look angles
        elevation, azimuth, slant_range_m = pointing.look_angles(
            sat_long, sat_lat, sat_alt, args.rx_long, args.rx_lat,
            args.rx_height, args.ref_ellipsoid)

        # Polarization skew
        #
        # NOTE: this parameter is used in the atmospheric loss model. It is
        # assumed equal to 45 degrees for circular polarization, according to
        # Recommendation ITU-R P.838-3.
        if (args.polarization == 'linear'):
            pol_skew = pointing.polarization_angle(sat_long, sat_lat,
                                                   args.rx_long, args.rx_lat)
        else:
            pol_skew = 45
    else:
        elevation = azimuth = pol_skew = None
        slant_range_m = args.slant_range * 1e3  # km to m

    if (elevation is not None and elevation < 0):
        el_label = "Object" if args.radar else "Satellite"
        logging.error("{} below the horizon".format(el_label))
        return

    # -------- Tx dish gain --------
    # Skipped if the EIRP is given directly.
    if (args.eirp is None):
        if args.tx_dish_gain is None:
            tx_dish = Antenna(freq=args.freq,
                              diameter=args.tx_dish_size,
                              efficiency=args.tx_dish_efficiency,
                              label="Tx dish")
        else:
            tx_dish = Antenna(freq=args.freq,
                              gain=args.tx_dish_gain,
                              efficiency=args.tx_dish_efficiency,
                              label="Tx dish")

    # -------- Rx dish gain --------
    if (args.rx_dish_gain is None):
        rx_dish = Antenna(freq=args.freq,
                          diameter=args.rx_dish_size,
                          efficiency=args.rx_dish_efficiency,
                          label="Rx dish")
    else:
        rx_dish = Antenna(freq=args.freq,
                          gain=args.rx_dish_gain,
                          efficiency=args.rx_dish_efficiency,
                          label="Rx dish")

    # -------- EIRP --------
    if (args.eirp is None):
        eirp_dbw = calc.eirp(args.tx_power, tx_dish.gain_db)
        util.log_result(
            "Tx Power",
            "{}".format(util.format_power(util.db_to_lin(args.tx_power))))
    else:
        eirp_dbw = args.eirp

    # The EIRP computed above could refer to an entire
    # transponder. Furthermore, it could refer to the power level in a given
    # direction when the amplifier operates in saturation. The following
    # function converts the EIRP computed above to the actual carrier EIRP.
    eirp_dbw = calc.carrier_eirp(eirp_dbw,
                                 args.obo,
                                 peb=args.carrier_peb,
                                 tp_bw=args.tp_bw)

    util.log_result("EIRP", "{:.2f} dBW".format(eirp_dbw))

    # -------- Path loss --------
    if (args.radar):
        radar_obj_gain_db = calc.radar_obj_gain(args.freq,
                                                args.radar_cross_section)
    else:
        radar_obj_gain_db = None
    path_loss_db = calc.path_loss(slant_range_m, args.freq, args.radar,
                                  radar_obj_gain_db)
    # TODO support bistatic radar mode. The path_loss function already supports
    # it. We just need to add --tx-lat/--tx-long options to separate the Tx and
    # Rx locations for bistatic mode.

    # -------- Atmospheric loss --------
    #
    # When the atmospheric loss is defined on argument "--atmospheric-loss", it
    # is prioritized over the atmospheric loss models. In contrast, when it is
    # not defined (the default), as long as the Rx station's coordinates and
    # the elevation are known, the atmospheric loss is computed from ITU-R
    # models. If the Rx and satellite coordinates are not available, the
    # atmospheric loss must be provided, otherwise the parser throws an error.
    #
    # Note that, ultimately, the atmospheric loss has two effects. It defines
    # the antenna noise temperature on reception (i.e., adds noise), and it
    # increases the path loss (i.e., reduces the signal power). In other words,
    # it disturbs both the numerator and the denominator of the C/N.
    if (args.atmospheric_loss is not None):
        one_way_atmospheric_loss_db = args.atmospheric_loss
    elif (all(v is not None for v in [args.rx_lat, args.rx_lat, elevation])):
        # Model the atmospheric attenuation for a given link availability.
        one_way_atmospheric_loss_db = propagation.atmospheric_attenuation(
            args.rx_lat, args.rx_long, elevation, pol_skew, args.freq,
            args.availability, rx_dish.diameter, rx_dish.aperture_efficiency)

    # On radar systems, assume the atmospheric loss is experienced twice.
    if (args.radar):
        two_way_atmospheric_loss_db = 2 * one_way_atmospheric_loss_db
        util.log_result("Total one-way atmospheric loss",
                        "{:.2f} dB".format(one_way_atmospheric_loss_db))
        util.log_result("Total two-way atmospheric loss",
                        "{:.2f} dB".format(two_way_atmospheric_loss_db))
    else:
        util.log_result("Total atmospheric loss",
                        "{:.2f} dB".format(one_way_atmospheric_loss_db))

    total_atmospheric_loss_db = two_way_atmospheric_loss_db if (args.radar) \
        else one_way_atmospheric_loss_db

    # -------- Reflected EIRP (radar mode only) --------
    # Compute the equivalent EIRP reflected off the radar object
    if (args.radar):
        one_way_path_loss_db = calc._path_loss(slant_range_m, args.freq)
        reflected_eirp_dbw = eirp_dbw \
            - one_way_path_loss_db \
            - one_way_atmospheric_loss_db \
            + radar_obj_gain_db
        util.log_result("Reflected EIRP",
                        "{:.2f} dBW".format(reflected_eirp_dbw))

    # -------- Noise figure --------
    coax_loss_db, coax_noise_fig_db = calc.coax_loss_nf(args.coax_length)

    if (args.lnb_noise_fig is None):
        lnb_noise_fig = calc.noise_temp_to_noise_fig(args.lnb_noise_temp)
    else:
        lnb_noise_fig = args.lnb_noise_fig

    util.log_result("LNB noise figure", "{:.2f} dB".format(lnb_noise_fig))

    noise_fig_db = calc.cascaded_noise_figure(
        [lnb_noise_fig, coax_noise_fig_db, args.rx_noise_fig],
        [args.lnb_gain, -coax_loss_db])

    # -------- System noise temperature --------
    effective_input_noise_temp = calc.noise_fig_to_noise_temp(noise_fig_db)

    util.log_result("Input-noise temperature",
                    "{:.2f} K".format(effective_input_noise_temp))

    if (args.antenna_noise_temp is None):
        antenna_noise_temp = calc.antenna_noise_temp(
            one_way_atmospheric_loss_db)
        # NOTE: consider the atmospheric loss only once here even if analyzing
        # a radar system. This call determines the sky's contribution to the Rx
        # antenna noise temperature.
    else:
        antenna_noise_temp = args.antenna_noise_temp

    T_syst = calc.rx_sys_noise_temp(antenna_noise_temp,
                                    effective_input_noise_temp,
                                    args.lna_feed_loss)
    T_syst_db = util.lin_to_db(T_syst)  # in dBK (for T_syst in K)

    # -------- Received Power and Flux Density --------
    if (args.radar):
        rx_flux_dbw_m2 = calc.rx_flux_density(reflected_eirp_dbw,
                                              slant_range_m,
                                              one_way_atmospheric_loss_db)
        # TODO: use the radar-to-Rx distance instead of the Tx-to-radar slant
        # range when running in bistatic radar mode (when supported).
    else:
        assert (one_way_atmospheric_loss_db == total_atmospheric_loss_db)
        rx_flux_dbw_m2 = calc.rx_flux_density(eirp_dbw, slant_range_m,
                                              one_way_atmospheric_loss_db)
    P_rx_dbw = calc.rx_power(eirp_dbw, path_loss_db, rx_dish.gain_db,
                             total_atmospheric_loss_db, args.mispointing_loss,
                             args.lna_feed_loss)

    # -------- Intermediate frequency (IF) Power --------
    P_if_dbw = P_rx_dbw + args.lnb_gain
    util.log_result("IF Power (LNB Output)",
                    "{:.2f} dBm".format(util.dbw_to_dbm(P_if_dbw)))

    # -------- Noise Power --------
    N_dbw = calc.noise_power(
        T_syst_db,
        args.bw,
    )

    # -------- Signal and Noise Power Spectral Densities --------
    sig_psd_dbw_hz = calc.spectral_density(P_rx_dbw,
                                           args.bw,
                                           label="Rx signal")
    noise_psd_dbw_hz = calc.spectral_density(N_dbw, args.bw, label="Noise")

    # -------- G/T and CNR --------
    g_over_t_db = calc.g_over_t(rx_dish.gain_db, T_syst_db)
    cnr_db = calc.cnr(P_rx_dbw, N_dbw)

    # -------- ASI and C/N+I --------
    if (args.asi):
        ci_db = calc.carrier_to_asi_ratio(rx_dish, args.asi_long_separation,
                                          args.asi_eirp_ratio)
        cnir_db = calc.cnir(cnr_db, ci_db)
    else:
        cnir_db = cnr_db
        ci_db = None

    # -------- Capacity --------
    capacity = calc.capacity(cnir_db, args.bw)

    # Results
    res = {
        'pointing': {
            'elevation': elevation,
            'azimuth': azimuth,
            'polarization_skew': pol_skew,
            'slant_range': slant_range_m
        },
        'eirp_db': eirp_dbw,
        'path_loss_db': path_loss_db,
        'atmospheric_loss_db': total_atmospheric_loss_db,
        'rx_dish_gain_db': rx_dish.gain_db,
        'noise_fig_db': {
            'lnb': lnb_noise_fig,
            'coax': coax_noise_fig_db,
            'total': noise_fig_db
        },
        'noise_temp_k': {
            'effective_input': effective_input_noise_temp,
            'system': T_syst
        },
        'power_dbw': {
            'carrier': P_rx_dbw,
            'noise': N_dbw,
            'if': P_if_dbw
        },
        'psd_dbw_hz': {
            'carrier': sig_psd_dbw_hz,
            'noise': noise_psd_dbw_hz
        },
        'rx_flux_dbw_m2': rx_flux_dbw_m2,
        'g_over_t_db': g_over_t_db,
        'cnr_db': cnr_db,
        'ci_db': ci_db,
        'cnir_db': cnir_db,
        'capacity_bps': capacity
    }

    # -------- Link Margin --------
    if (args.min_cnr is not None):
        effective_min_cnr = args.min_cnr + args.impl_margin
        margin_db = cnir_db - effective_min_cnr
        util.log_result("Link margin", "{:.2f} dB".format(margin_db))
        res['margin_db'] = margin_db

    if (verbose and args.json):
        print(json.dumps(res, indent=4, ensure_ascii=True))

    return res


def main():
    parser = get_parser()
    args = parser.parse_args()
    if (not args.json):
        configure_logging()
    validate(parser, args)
    analyze(args, verbose=True)



"""Satellite look angle computations

References:

- [1] https://www.ngs.noaa.gov/CORS/Articles/SolerEisemannJSE.pdf.
- [2] https://en.wikipedia.org/wiki/Earth_radius.
- [3] Maral, Gerard., Sun, Zhili., Bousquet, Michel. Satellite Communications
  Systems: Systems, Techniques and Technology. 3rd ed.

"""
import logging
import os
from datetime import datetime
from math import sqrt, sin, cos, tan, atan, atan2, degrees, radians
from urllib.parse import quote

import numpy as np
from skyfield.api import load, wgs84

from . import util

CELESTRAK_GROUPS = [
    'active', 'stations', 'geo', 'intelsat', 'ses', 'iridium', 'iridium-NEXT',
    'starlink', 'oneweb', 'orbcomm', 'globalstar', 'swarm', 'amateur',
    'x-comm', 'other-comm', 'satnogs', 'gorizont', 'raduga', 'molniya',
    'weather', 'noaa', 'goes', 'resource', 'sarsat', 'dmc', 'tdrss', 'argos',
    'planet', 'spire', 'gnss', 'gps-ops', 'glo-ops', 'galileo', 'beidou',
    'sbas', 'nnss', 'musson', 'military', 'radar', 'cubesat', 'other'
]
CELESTRAK_BASE_URL = "https://celestrak.org/NORAD/elements/gp.php"
skyfield_ts = load.timescale()


def get_default_tle_dataset_dir():
    """Get the default directory for saving TLE datasets"""
    return os.path.join(util.get_default_lb_dir(), "tle")


def get_sat_pos_by_tle(name,
                       group=None,
                       obs_time=None,
                       save_dir=None,
                       no_save=False):
    """Get satellite position from Two-Line Element (TLE) predictions

    First, searches for the satellite on CelesTrak's TLE database. Then,
    computes the expected satellite position on the given observation time
    using the implementation from the Skyfield package.

    The TLE dataset is queried using the format described in
    https://celestrak.org/NORAD/documentation/gp-data-formats.php. More
    specifically, it is either requested with:

    https://celestrak.org/NORAD/elements/gp.php?NAME=VALUE&FORMAT=tle

    or

    https://celestrak.org/NORAD/elements/gp.php?GROUP=VALUE&FORMAT=tle

    The former is used when the satellite is specified by name only (group set
    to None). The second format is used when the specific group on CelesTrak's
    database is also informed. See the groups at
    https://celestrak.org/NORAD/elements/. For example, the 'active' group
    holds a long list of active satellites, whereas the 'geo' group focuses on
    active geosynchronous satellites, and so on.

    This function downloads the TLE dataset before processing, and the
    downloads are saved as txt files at `~/.link-budget/tle/` by default. When
    the dataset is already available locally (downloaded previously), this
    function does not need to download it again. This behavior can be
    customized using the `save_dir` and `no_save` arguments.

    Args:
        name (str): Satellite name on the TLE database.
        group (str, optional): Group to which the satellite belongs on the
            CelesTrak database. Defaults to None, in which case the satellite
            is searched by name over the entire database.
        obs_time (datetime, optional): Observation time. Defaults to
            None, in which case the current time is used.
        save_dir (str, optional): Directory on which the downloaded TLE dataset
            should be saved. Defaults to None, in which case the
            `~/.link-budget/tle/` directory is used.
        no_save (bool, optional): Whether to save the downloaded TLE dataset
            permanently on the save directory (save_dir) to speed up future
            queries.

    Returns:
        tuple: Satellite longitude (degrees), latitude (degrees), and altitude
        (meters).
    """
    name = name.upper()  # ensure upper case

    if save_dir is None:
        save_dir = get_default_tle_dataset_dir()

    if not os.path.exists(save_dir):
        os.makedirs(save_dir)

    if group is not None:
        url = CELESTRAK_BASE_URL + "?GROUP={}&FORMAT=tle".format(group)
        filename = 'tle-group-{}.txt'.format(group)
    else:
        url = CELESTRAK_BASE_URL + "?NAME={}&FORMAT=tle".format(quote(name))
        filename = 'tle-name-{}.txt'.format(name.replace(" ", "-"))

    filename = os.path.join(save_dir, filename)
    previously_downloaded = os.path.exists(filename)
    satellites = {
        sat.name: sat
        for sat in load.tle_file(url, filename=filename)
    }

    if name not in satellites:
        logging.error(
            "Satellite \"{}\" not found on the TLE database.".format(name))
        closest_options = []
        for key in satellites.keys():
            if name in key:
                closest_options.append(key)
        if closest_options:
            logging.info("Perhaps you mean one of the following options?")
            for option in closest_options:
                logging.info("- " + option)
        if no_save and not previously_downloaded:  # just downloaded
            os.remove(filename)
        raise ValueError("Invalid satellite name")

    satellite = satellites[name]

    obs_time = datetime.utcnow() if obs_time is None else obs_time
    t = skyfield_ts.utc(obs_time.year, obs_time.month, obs_time.day,
                        obs_time.hour, obs_time.minute, obs_time.second)
    geocentric = satellite.at(t)
    lat, lon = wgs84.latlon_of(geocentric)
    height = wgs84.height_of(geocentric)

    util.log_result(
        "Satellite Pos (Lat, Lon, Alt)",
        "{:.2f}°, {:.2f}°, {:.2f} km".format(lat.degrees, lon.degrees,
                                             height.km))

    if no_save and not previously_downloaded:  # just downloaded
        os.remove(filename)

    return lon.degrees, lat.degrees, height.m


def look_angles(sat_long,
                sat_lat,
                sat_alt,
                rx_long,
                rx_lat,
                rx_height=0,
                ellipsoid="WGS84"):
    """Calculate look angles (elevation, azimuth) and slant range

    Computes the angles relative to a reflector, either active (satellite) or
    passive (radar object). Compute using the rigorous ellipsoidal approach
    discussed in [1].

    Args:
        sat_long       : Subsatellite point's longitude in degrees.
        sat_lat:       : Subsatellite point's geodetic latitude in degrees.
        sat_alt        : Satellite/reflector altitude in meters above the
                         reference ellipsoid.
        rx_long        : Longitude of the receiver station in degrees.
        rx_lat         : Geodetic latitude of the receiver station in degrees.
        rx_height      : Receiver station's orthometric height (height above
                         sea-level) in meters. Defaults to zero.
        ellipsoid      : Reference ellipsoid for the computation, GRS80 or
                         WGS84. Defaults to WGS84.

    Note:
        Positive longitudes are east of the prime meridian and negative
        longitudes are west of the prime meridian. Positive latitudes are north
        of the equator and negative latitudes are south of the equator.

    Returns:
        Tuple with elevation (degrees), azimuth (degrees) and slant range (m).

    """

    # Step 1: PROGRAM INPUTS

    # Convert to radians
    sat_long = radians(sat_long)
    sat_lat = radians(sat_lat)
    rx_long = radians(rx_long)
    rx_lat = radians(rx_lat)

    # Step 2: TRANSFORMATION CURVILINEAR TO CARTESIAN COORDINATES

    # Ellipsoid parameters from GRS80
    f_inv = {
        'GRS80': 298.257222100882711,
        'WGS84': 298.257223563
    }  # reciprocal flattening
    f = 1 / f_inv[ellipsoid]  # flattening
    e_sq = 2 * f - f**2  # eccentricity squared

    # Earth parameters
    R_eq = 6378.137e3  # equatorial radius in meters (see [2])
    r = R_eq + sat_alt  # from the earth's center to the spacecraft

    # Principal radius of curvature in the prime vertical (see the the
    # discussion below Eq 12 in [1]). The computation is also discussed in [2].
    N = R_eq / sqrt(1 - e_sq * sin(rx_lat)**2)

    # Ellipsoidal (geodetic) height of the antenna location:
    Ng = 0  # undulation of the geoid or geoid height
    h = Ng + rx_height

    # Rectangular coordinates of the antenna location, using Eq. 12 from [1]:
    x_p = (N + h) * cos(rx_long) * cos(rx_lat)
    y_p = (N + h) * sin(rx_long) * cos(rx_lat)
    z_p = (N * (1 - e_sq) + h) * sin(rx_lat)

    # Rectangular coordinates of the satellite. See Fig. 5 in [1]:
    x_s = r * cos(sat_long) * cos(sat_lat)
    y_s = r * sin(sat_long) * cos(sat_lat)
    z_s = (N * (1 - e_sq) + sat_alt) * sin(sat_lat)

    # Step 3: SATELLITE COMPONENTS ON LOCAL (x, y, z) COORDINATES
    rect_coor = np.array([x_s, y_s, z_s]) - np.array([x_p, y_p, z_p])

    # rect_coor is a vector starting on the receiver position P and ending on
    # the satellite S (i.e., the topocentric range PS). In other words, it
    # represents the satellite rectangular coordinates referenced to the
    # receiver position. The Euclidean norm of the vector is the slant (or
    # topocentric) range:
    slant_range = np.linalg.norm(rect_coor)

    # Step 4: SATELLITE COMPONENTS ON LOCAL (e, n, u)
    #
    # e-axis points to (geodetic) east; n to (geodetic) north; and u to
    # (geodetic) zenith.

    # Rotation matrix - Eq. 9b from [1]:
    rot_mtx = np.array(
        [[-sin(rx_long), cos(rx_long), 0],
         [
             -sin(rx_lat) * cos(rx_long), -sin(rx_lat) * sin(rx_long),
             cos(rx_lat)
         ],
         [cos(rx_lat) * cos(rx_long),
          cos(rx_lat) * sin(rx_long),
          sin(rx_lat)]])
    # Conversion using Eq. 10 [1]:
    geodetic_coor = np.dot(rot_mtx, rect_coor)

    # Step 5: GEODETIC AZIMUTH AND GEODETIC VERTICAL ANGLE
    e, n, u = geodetic_coor
    azimuth = atan2(e, n)
    vert_angle = atan(u / sqrt(e**2 + n**2))  # elevation

    azimuth_degrees = degrees(azimuth) % 360
    elevation_degrees = degrees(vert_angle)

    util.log_result("Elevation", "{:.2f} degrees".format(elevation_degrees))
    util.log_result("Azimuth", "{:.2f} degrees".format(azimuth_degrees))
    util.log_result("Distance", "{:.2f} km".format(slant_range / 1e3))

    return elevation_degrees, azimuth_degrees, slant_range


def polarization_angle(sat_long, sat_lat, rx_long, rx_lat):
    """Compute the polarization angle (skew) at a given location

    A linearly polarized satellite transmission has the electric field oriented
    at a constant angle relative to a reference plane. For a satellite antenna,
    the reference plane is the equatorial plane. If the polarization is
    horizontal, the electric field is parallel to the equatorial
    plane. Otherwise, when the linear polarization is vertical, it is
    perpendicular to the equatorial plane.

    The earth station antenna feed must have its polarization aligned with the
    polarization plane of the received wave. However, the local vertical
    (normal) plane does not match the equatorial plane due to the curvature of
    the earth. Hence, there is an angle difference between the polarization of
    the signal transmitted by the satellite and the apparent polarization of
    the received signal, which is known as the polarization angle or
    polarization skew (sometimes also referred to as "polarity", "polarity
    skew", "LNB skew", or just "skew").

    When pointing an linearly-polarized earth station antenna, this angle has
    to be taken into account. In contrast, if pointing a circularly-polarized
    antenna, there is no need to compensate for the polarization skew.

    For geostationary satellites, [3] presents the skew formula that follows:

    .. math::

        \\psi = \\tan^{-1} \\left( \\frac{\\sin(L)}{\\tan(l)} \\right)

    where :math:`L` is the relative longitude (difference between the earth
    station's longitude and the satellite longitude) and :math:`l` is the earth
    station's latitude.

    As indicated by the equation, the skew is 0 when the earth station and the
    satellite are at the same longitude (when the relative longitude is zero).

    Note that, for a positive skew value, the LNB must be rotated clockwise,
    whereas, for a negative polarization angle, the LNB must be rotated
    counterclockwise. Furthermore, note that the clockwise and counterclockwise
    directions are relative to the front face (the feed horn) of the LNB. In
    other words, for a positive skew value, someone standing behind the dish
    and facing the satellite in the sky would rotate the LNB clockwise.

    Args:
        sat_long (float): Subsatellite point's longitude in degrees.
        sat_lat  (float): Subsatellite point's geodetic latitude in degrees.
        rx_long  (float): Earth station's longitude in degrees.
        rx_lat   (float): Earth station's geodetic latitude in degrees.

    Returns:
        float: Polarization angle in degrees.

    """
    # Convert to radians
    sat_long = radians(sat_long)
    sat_lat = radians(sat_lat)
    rx_long = radians(rx_long)
    rx_lat = radians(rx_lat)

    # FIXME support computation for satellites in non-geostationary orbit
    if sat_lat != 0:
        logging.warning(
            "The polarization angle is inaccurate for non-geostationary "
            "satellites")

    # Computation based on Eq. (8.22c) from [3]:
    delta_long = rx_long - sat_long
    pol_angle_rad = atan(sin(delta_long) / tan(rx_lat))

    # Return the polarization angle in degrees
    pol_angle = degrees(pol_angle_rad)
    util.log_result("Polarizationa angle", "{:.2f} degrees".format(pol_angle))
    return pol_angle



import unittest

from . import util


class TestUtil(unittest.TestCase):

    def test_format_rate(self):
        self.assertEqual(util.format_rate(1e-1), "0.10 bps")
        self.assertEqual(util.format_rate(1e0), "1.00 bps")
        self.assertEqual(util.format_rate(1e1), "10.00 bps")
        self.assertEqual(util.format_rate(1e2), "100.00 bps")
        self.assertEqual(util.format_rate(1e3), "1.00 kbps")
        self.assertEqual(util.format_rate(1e4), "10.00 kbps")
        self.assertEqual(util.format_rate(1e5), "100.00 kbps")
        self.assertEqual(util.format_rate(1e6), "1.00 Mbps")
        self.assertEqual(util.format_rate(1e7), "10.00 Mbps")
        self.assertEqual(util.format_rate(1e8), "100.00 Mbps")
        self.assertEqual(util.format_rate(1e9), "1.00 Gbps")
        self.assertEqual(util.format_rate(1e10), "10.00 Gbps")

    def test_format_area(self):
        self.assertEqual(util.format_area(1e1), "10.00 m2")
        self.assertEqual(util.format_area(1e0), "1.00 m2")
        self.assertEqual(util.format_area(1e-1), "10.00 dm2")
        self.assertEqual(util.format_area(1e-2), "1.00 dm2")
        self.assertEqual(util.format_area(1e-3), "10.00 cm2")
        self.assertEqual(util.format_area(1e-4), "1.00 cm2")
        self.assertEqual(util.format_area(1e-5), "0.10 cm2")

    def test_format_power(self):
        self.assertEqual(util.format_power(1e4), "10.00 kW")
        self.assertEqual(util.format_power(1e3), "1.00 kW")
        self.assertEqual(util.format_power(1e2), "100.00 W")
        self.assertEqual(util.format_power(1e1), "10.00 W")
        self.assertEqual(util.format_power(1e0), "1.00 W")
        self.assertEqual(util.format_power(1e-1), "100.00 mW")
        self.assertEqual(util.format_power(1e-2), "10.00 mW")
        self.assertEqual(util.format_power(1e-3), "1.00 mW")
        self.assertEqual(util.format_power(1e-4), "0.10 mW")



import itur

from . import util


def atmospheric_attenuation(lat, lng, elevation, pol_skew, freq, availability,
                            dish_size, dish_eff):
    """Total attenuation due to multiple sources of atmospheric attenuation

    A wrapper to obtain the total atmospheric attenuation through ITU-Rpy.

    Args:
        lat (float): Earth station's latitude in degrees.
        lng (float): Earth station's longitude in degrees.
        elevation (float): Elevation angle in degrees.
        freq (float): Carrier frequency in Hz.
        availability (float): Target link availability within [95 to 99.999%].
        dish_size (float): Earth station's dish diameter in m.
        dish_eff (float): Dish aperture efficiency within [0, 1].

    Returns:
        float: Total atmospheric attenuation in dB.

    """
    f_ghz = freq / 1e9
    unavailability = 100 - availability
    a_g, a_c, a_r, a_s, a_t = itur.atmospheric_attenuation_slant_path(
        lat,
        lng,
        f_ghz,
        elevation,
        unavailability,
        dish_size,
        eta=dish_eff,
        tau=pol_skew,
        return_contributions=True)
    util.log_result("Rain attenuation @ {:g}%".format(unavailability),
                    "{:.2f}".format(a_r))
    util.log_result("Cloud attenuation @ {:g}%".format(unavailability),
                    "{:.2f}".format(a_c))
    util.log_result("Gaseous attenuation @ {:g}%".format(unavailability),
                    "{:.2f}".format(a_g))
    util.log_result("Tropo. scintillation @ {:g}%".format(unavailability),
                    "{:.2f}".format(a_s))
    return a_t.value



import os
import shutil
import unittest

from skyfield.api import load, wgs84

from . import pointing
from .constants import GEOSYNC_ORBIT


class TestPointing(unittest.TestCase):

    def test_gso_look_angles(self):
        """Test look angle calculations for GSO satellites

        Test Rx stations at all four quadrants (SW, NW, NE, SE). The look
        angles are compared to those obtained with the dish pointer tool
        (https://www.dishpointer.com).

        SW Station:
        1) Sao Paulo (Latitude: -23.5505°, Longitude: -46.6333°)
           Satellite: Eutelsat 113 (113° West)

        NW Stations:
        1) Washington DC (Latitude: 38.9072°, Longitude: -77.0369°)
           Satellite: Eutelsat 113 (113° West)
        2) Los Angeles (Latitude: 34.0522°, Longitude: -118.2437°)
           Satellite: Galaxy 18 (123° West)

        NE Station:
        1) Berlin (Latitude: 52.5200°, Longitude: 13.4050°)
           Satellite: Telstar 11N (37.5° West)

        SE Stations:
        1) Sydney (Latitude: -33.8688°, Longitude: 151.2093°)
           Satellite: Telstar 18V (138° East)
        2) Cape Town (Latitude: -33.9249°, Longitude: 18.4241°)
           Satellite: Telstar 11N (37.5° West)

        """
        setup_info = [
            {
                'sat_long': -113,
                'rx_lat': -23.5505,
                'rx_long': -46.6333,
                'elevation': 13.1,
                'azimuth': 279.9,
                'distance': 40260,
                'lnb_skew': -64.6
            },
            {
                'sat_long': -113,
                'rx_lat': 38.9072,
                'rx_long': -77.0369,
                'elevation': 31.6,
                'azimuth': 229.1,
                'distance': 38469,
                'lnb_skew': 36
            },
            {
                'sat_long': -123,
                'rx_lat': 34.0522,
                'rx_long': -118.2437,
                'elevation': 50.1,
                'azimuth': 188.4,
                'distance': 37077,
                'lnb_skew': 7
            },
            {
                'sat_long': -37.5,
                'rx_lat': 52.5200,
                'rx_long': 13.4050,
                'elevation': 14.1,
                'azimuth': 237.2,
                'distance': 40151,
                'lnb_skew': 30.8
            },
            {
                'sat_long': 138,
                'rx_lat': -33.8688,
                'rx_long': 151.2093,
                'elevation': 48.1,
                'azimuth': 337.2,
                'distance': 37202,
                'lnb_skew': -18.8
            },
            {
                'sat_long': -37.5,
                'rx_lat': -33.9249,
                'rx_long': 18.4241,
                'elevation': 19.5,
                'azimuth': 290.7,
                'distance': 39604,
                'lnb_skew': -50.9
            },
        ]

        # Geostationary latitude and altitude
        sat_lat = 0
        sat_alt = GEOSYNC_ORBIT

        for info in setup_info:
            elevation, azimuth, slant_range = pointing.look_angles(
                info['sat_long'], sat_lat, sat_alt, info['rx_long'],
                info['rx_lat'])
            slant_range_km = slant_range / 1e3
            pol_angle = pointing.polarization_angle(info['sat_long'], sat_lat,
                                                    info['rx_long'],
                                                    info['rx_lat'])

            self.assertAlmostEqual(elevation, info['elevation'], delta=0.1)
            self.assertAlmostEqual(azimuth, info['azimuth'], delta=0.1)
            # The distance sometimes differs by a few km. Maybe dishpointer
            # considers the actual longitude of the satellite based on
            # ephemeris data instead of the nominal.
            self.assertAlmostEqual(slant_range_km, info['distance'], delta=12)
            self.assertAlmostEqual(pol_angle, info['lnb_skew'], delta=0.05)

    def test_ngso_look_angles(self):
        """Test look angle calculations for NGSO satellites

        Tested by comparison to the results obtained with Skyfield.

        """
        # Rx in Washington DC
        rx_lat = 38.9072
        rx_long = -77.0369
        rx_height = 0
        observer_pos = wgs84.latlon(rx_lat, rx_long, rx_height)

        # NGSO Satellite: Starlink-3699 covering DC at 2023-03-17T18:31:45
        ts = load.timescale()
        sat_long, sat_lat, sat_alt = pointing.get_sat_pos_by_tle(
            'STARLINK-3699')
        t = ts.utc(2023, 3, 17, 18, 31, 45)
        sat_pos = wgs84.latlon(sat_lat, sat_long, sat_alt)

        difference = sat_pos - observer_pos
        topocentric = difference.at(t)
        skyfield_el, skyfield_az, _ = topocentric.altaz()

        el, az, _ = pointing.look_angles(sat_long, sat_lat, sat_alt, rx_long,
                                         rx_lat, rx_height)

        self.assertAlmostEqual(el, skyfield_el.degrees, delta=0.25)
        self.assertAlmostEqual(az, skyfield_az.degrees, delta=0.25)
        # TODO: the slant range differs significantly. Understand why.

    def test_get_sat_pos_by_tle(self):
        """Test satellite position through TLE-based prediction"""

        # Basic query using the default download directory
        lon, lat, alt = pointing.get_sat_pos_by_tle('DIRECTV 9S')
        self.assertAlmostEqual(lon, -101, delta=0.3)
        self.assertAlmostEqual(lat, 0, delta=0.1)
        self.assertAlmostEqual(alt, GEOSYNC_ORBIT, delta=20e3)

        # The dataset should be saved
        expected_file = os.path.join(pointing.get_default_tle_dataset_dir(),
                                     'tle-name-DIRECTV-9S.txt')
        self.assertTrue(os.path.exists(expected_file))
        os.remove(expected_file)

        # Same thing but specifying another download directory
        tle_download_dir = '/tmp/link-budget-test/'
        lon, lat, alt = pointing.get_sat_pos_by_tle('DIRECTV 9S',
                                                    save_dir=tle_download_dir)
        expected_file = os.path.join(tle_download_dir,
                                     'tle-name-DIRECTV-9S.txt')
        self.assertTrue(os.path.exists(expected_file))

        # If the no_save option is specified but the dataset is already
        # available, the dataset must be kept (not removed in the end)
        lon, lat, alt = pointing.get_sat_pos_by_tle('DIRECTV 9S',
                                                    save_dir=tle_download_dir,
                                                    no_save=True)
        self.assertTrue(os.path.exists(expected_file))

        # On the other hand, if no_save is True and the dataset does not exist
        # yet, it should be removed in the end
        os.remove(expected_file)  # pretend the following is the first download
        lon, lat, alt = pointing.get_sat_pos_by_tle('DIRECTV 9S',
                                                    save_dir=tle_download_dir,
                                                    no_save=True)
        self.assertFalse(os.path.exists(expected_file))

        # Query under a specific group
        lon, lat, alt = pointing.get_sat_pos_by_tle('DIRECTV 9S',
                                                    group='active',
                                                    save_dir=tle_download_dir)
        self.assertAlmostEqual(lon, -101, delta=0.3)
        self.assertAlmostEqual(lat, 0, delta=0.1)
        self.assertAlmostEqual(alt, GEOSYNC_ORBIT, delta=20e3)

        # The group query downloads the entire group's dataset (prefixed with
        # 'tle-group'), unlike the name-only query, which downloads the TLE
        # entries matching the given name (saved with 'tle-name' prefix).
        expected_file = os.path.join(tle_download_dir, 'tle-group-active.txt')
        self.assertTrue(os.path.exists(expected_file))

        # Invalid satellite name
        with self.assertRaises(ValueError):
            pointing.get_sat_pos_by_tle('EUTELSAT 113',
                                        group='active',
                                        save_dir=tle_download_dir)

        # The same saving behavior must apply even when the satellite is not
        # found. In the above command, the active group's TLE dataset will be
        # downloaded even though the satellite name is invalid. If we add the
        # no_save option, the dataset should be removed in the end, except if
        # the dataset already existed before due to a previous query.
        with self.assertRaises(ValueError):
            pointing.get_sat_pos_by_tle('EUTELSAT 113',
                                        group='active',
                                        save_dir=tle_download_dir,
                                        no_save=True)
        self.assertTrue(
            os.path.exists(expected_file))  # still there (existed before)

        os.remove(expected_file)  # pretend the following is the first download
        with self.assertRaises(ValueError):
            pointing.get_sat_pos_by_tle('EUTELSAT 113',
                                        group='active',
                                        save_dir=tle_download_dir,
                                        no_save=True)
        self.assertFalse(os.path.exists(expected_file))  # removed in the end

        shutil.rmtree(tle_download_dir)



import re
import sys
from pathlib import Path
from setuptools import setup, find_packages

if sys.version_info[0] < 3:
    raise SystemExit("Error: the link-budget tool requires Python 3")

version = re.search(r'^__version__\s*=\s*"(.*)"',
                    open('linkbudget/cli.py').read(), re.M).group(1)

long_description = (Path(__file__).parent / "README.md").read_text()

setup(name="link-budget",
      packages=find_packages(),
      entry_points={"console_scripts": ['link-budget = linkbudget.cli:main']},
      version=version,
      description="Link budget calculator for satellite and radar systems",
      long_description=long_description,
      long_description_content_type='text/markdown',
      author="Igor Freire",
      author_email="igor@blockstream.com",
      url="https://github.com/igorauad/link-budget",
      classifiers=[
          'Programming Language :: Python :: 3',
          "Programming Language :: Python :: 3 :: Only",
          "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"
      ],
      install_requires=[
          "itur>=0.4.0",
          "numpy>=1.19.4",
          "python-dateutil>=2.7.0",
          "scipy>=1.7",
          "skyfield==1.42",
      ],
      python_requires='>=3')
